# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class SessionEvent(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'event_type': 'str',
        'log_level': 'LogLevelTypes',
        'timestamp': 'datetime',
        'description': 'str',
        'cause': 'SessionEventCause',
        'id': 'str'
    }

    attribute_map = {
        'event_type': 'event_type',
        'log_level': 'log_level',
        'timestamp': 'timestamp',
        'description': 'description',
        'cause': 'cause',
        'id': 'id'
    }

    def __init__(self, event_type=None, log_level=None, timestamp=None, description=None, cause=None, id=None, local_vars_configuration=None):  # noqa: E501
        """SessionEvent - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._event_type = None
        self._log_level = None
        self._timestamp = None
        self._description = None
        self._cause = None
        self._id = None
        self.discriminator = None

        self.event_type = event_type
        self.log_level = log_level
        self.timestamp = timestamp
        self.description = description
        if cause is not None:
            self.cause = cause
        self.id = id

    @property
    def event_type(self):
        """Gets the event_type of this SessionEvent.  # noqa: E501

        The category of session event.  # noqa: E501

        :return: The event_type of this SessionEvent.  # noqa: E501
        :rtype: str
        """
        return self._event_type

    @event_type.setter
    def event_type(self, event_type):
        """Sets the event_type of this SessionEvent.

        The category of session event.  # noqa: E501

        :param event_type: The event_type of this SessionEvent.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and event_type is None:  # noqa: E501
            raise ValueError("Invalid value for `event_type`, must not be `None`")  # noqa: E501

        self._event_type = event_type

    @property
    def log_level(self):
        """Gets the log_level of this SessionEvent.  # noqa: E501

        The severity of the session event.  # noqa: E501

        :return: The log_level of this SessionEvent.  # noqa: E501
        :rtype: LogLevelTypes
        """
        return self._log_level

    @log_level.setter
    def log_level(self, log_level):
        """Sets the log_level of this SessionEvent.

        The severity of the session event.  # noqa: E501

        :param log_level: The log_level of this SessionEvent.  # noqa: E501
        :type: LogLevelTypes
        """
        if self.local_vars_configuration.client_side_validation and log_level is None:  # noqa: E501
            raise ValueError("Invalid value for `log_level`, must not be `None`")  # noqa: E501

        self._log_level = log_level

    @property
    def timestamp(self):
        """Gets the timestamp of this SessionEvent.  # noqa: E501

        The time at which the session event occurred.  # noqa: E501

        :return: The timestamp of this SessionEvent.  # noqa: E501
        :rtype: datetime
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp):
        """Sets the timestamp of this SessionEvent.

        The time at which the session event occurred.  # noqa: E501

        :param timestamp: The timestamp of this SessionEvent.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and timestamp is None:  # noqa: E501
            raise ValueError("Invalid value for `timestamp`, must not be `None`")  # noqa: E501

        self._timestamp = timestamp

    @property
    def description(self):
        """Gets the description of this SessionEvent.  # noqa: E501

        A human readable description of the session event.  # noqa: E501

        :return: The description of this SessionEvent.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this SessionEvent.

        A human readable description of the session event.  # noqa: E501

        :param description: The description of this SessionEvent.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and description is None:  # noqa: E501
            raise ValueError("Invalid value for `description`, must not be `None`")  # noqa: E501

        self._description = description

    @property
    def cause(self):
        """Gets the cause of this SessionEvent.  # noqa: E501

        The reason why the session event occurred. None indicates an unknown cause.  # noqa: E501

        :return: The cause of this SessionEvent.  # noqa: E501
        :rtype: SessionEventCause
        """
        return self._cause

    @cause.setter
    def cause(self, cause):
        """Sets the cause of this SessionEvent.

        The reason why the session event occurred. None indicates an unknown cause.  # noqa: E501

        :param cause: The cause of this SessionEvent.  # noqa: E501
        :type: SessionEventCause
        """

        self._cause = cause

    @property
    def id(self):
        """Gets the id of this SessionEvent.  # noqa: E501

        A unique identifier for the session event.  # noqa: E501

        :return: The id of this SessionEvent.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this SessionEvent.

        A unique identifier for the session event.  # noqa: E501

        :param id: The id of this SessionEvent.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SessionEvent):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SessionEvent):
            return True

        return self.to_dict() != other.to_dict()



from typing import Tuple
from unittest.mock import Mock

import pytest

from anyscale.api_utils.exceptions.job_errors import NoJobRunError
from anyscale.api_utils.job_util import _get_job_run_id
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as BaseApi
from tests.api_utils.conftest import get_mock_base_api


def test_get_job_run_id():
    job_id = "prodjob_123"
    job_run_id = "job_123"
    # Returns from job_run_id
    assert _get_job_run_id(Mock(), job_run_id=job_run_id) == job_run_id

    # Returns from ProductionJob.last_job_run_id of job_id
    mock_prod_job = Mock(last_job_run_id=job_run_id)
    mock_base_api = get_mock_base_api(
        {BaseApi.get_production_job.__name__: Mock(result=mock_prod_job)}
    )
    assert _get_job_run_id(mock_base_api, job_id=job_id) == job_run_id

    # Raises NoJobRunError if ProductionJob.last_job_run_id is None
    mock_base_api.get_production_job.return_value.result.last_job_run_id = None
    with pytest.raises(
        NoJobRunError, match="Production job prodjob_123 has no job runs."
    ):
        _get_job_run_id(mock_base_api, job_id="prodjob_123")


@pytest.mark.parametrize("id_args", [(None, None), ("prodjob_123", "job_123")])
def test_get_job_run_id_errors_if_not_only_one_id_arg_provided(
    id_args: Tuple[str, str]
):
    job_id, job_run_id = id_args
    with pytest.raises(
        AssertionError,
        match="Exactly one of `job_id` or `job_run_id` must be provided.",
    ):
        _get_job_run_id(Mock(), job_id=job_id, job_run_id=job_run_id)



from io import StringIO
from json import dumps as json_dumps
import pathlib
from typing import Optional

import click
import yaml

import anyscale
from anyscale.cli_logger import BlockLogger
from anyscale.controllers.schedule_controller import ScheduleController
from anyscale.schedule.models import ScheduleConfig, ScheduleState


log = BlockLogger()  # CLI Logger


@click.group("schedule", help="Create and manage Anyscale Schedules.")
def schedule_cli() -> None:
    pass


def _read_identifiers_from_config_file(path: str):
    """Read the 'name', 'cloud', and 'project' properties from the config file at `path`.

    Return the identifers as a ScheduleIdentifiers object.
    """
    if not pathlib.Path(path).is_file():
        raise click.ClickException(f"Config file not found at path: '{path}'.")

    with open(path) as f:
        config = yaml.safe_load(f)

    if config is None or "job_config" not in config:
        raise click.ClickException(
            f"No 'job_config' property found in config file '{path}'."
        )

    job_config = config.get("job_config")
    name = job_config.get("name", None)
    cloud = job_config.get("cloud", None)
    project = job_config.get("project", None)

    return name, cloud, project


def _validate_schedule_identifiers(
    name: Optional[str], id: Optional[str], config_file: Optional[str]  # noqa: A002
):
    num_passed = sum(val is not None for val in [name, id, config_file])
    if num_passed == 0:
        raise click.ClickException(
            "One of '--name', '--id', or '--config-file' must be provided."
        )

    if num_passed > 1:
        raise click.ClickException(
            "Only one of '--name', '--id', and '--config-file' can be provided."
        )


@schedule_cli.command(name="create")
@click.argument("schedule_config_file", required=True)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option(
    "--description", required=False, default=None, help="Description of schedule."
)
def create(
    schedule_config_file: str, name: Optional[str], description: Optional[str],
) -> None:
    """ Create or Update a Schedule

    This function accepts 1 argument, a path to a YAML config file that defines this schedule.

    Note: if a schedule with the name exists in the specified project, it will be updated instead.
    """
    log.warning("DEPRECATED: Use `anyscale schedule apply` instead.")
    job_controller = ScheduleController()
    job_controller.apply(
        schedule_config_file, name, description,
    )


@schedule_cli.command(name="update")
@click.argument("schedule_config_file", required=True)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option(
    "--description", required=False, default=None, help="Description of schedule."
)
def update(
    schedule_config_file: str, name: Optional[str], description: Optional[str],
) -> None:
    """ Create or Update a Schedule

    This function accepts 1 argument, a path to a YAML config file that defines this schedule.
    """
    log.warning("DEPRECATED: Use `anyscale schedule apply` instead.")
    job_controller = ScheduleController()
    job_controller.apply(
        schedule_config_file, name, description,
    )


@schedule_cli.command(name="apply")
@click.option(
    "--config-file",
    "-f",
    required=True,
    type=str,
    help="Path to a YAML config file to use for this schedule. Command-line flags will overwrite values read from the file.",
)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
def apply(config_file: str, name: Optional[str],) -> None:
    """ Create or Update a Schedule

    The schedule should be specified in a YAML config file.
    """
    if not pathlib.Path(config_file).is_file():
        raise click.ClickException(f"Schedule config file '{config_file}' not found.")

    config = ScheduleConfig.from_yaml(config_file)

    if name is not None:
        config = config.options(job_config=config.job_config.options(name=name),)

    log.info(f"Applying schedule with config {config}.")
    anyscale.schedule.apply(config)


@schedule_cli.command(name="list",)
@click.option(
    "--name", "-n", required=False, default=None, help="Filter by the name of the job"
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
def list(  # noqa: A001
    name: Optional[str] = None, id: Optional[str] = None  # noqa: A002
) -> None:
    """ List Schedules

    You can optionally filter schedules by rowname.
    """
    job_controller = ScheduleController()
    job_controller.list(name=name, id=id)


@schedule_cli.command(name="pause")
@click.option(
    "--config-file",
    "-f",
    required=False,
    type=str,
    help="Path to a YAML config file to use for this schedule.",
)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The named Anyscale Cloud for the schedule. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the schedule. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
def pause(
    config_file: str, name: str, cloud: str, project: str, id: str  # noqa: A002
) -> None:
    """Pause a Schedule.

    You can pause a schedule by config file, name, or id.

    To specify the schedule by name, use the --name flag. You can specify the cloud with --cloud and the project with --project.

    To specify the schedule by id, use the --id flag.

    To specify the schedule by config file, use --config-file. Ensure that name and optionally cloud and project are specified in the
    config file's job config.
    """
    _validate_schedule_identifiers(name=name, id=id, config_file=config_file)

    if id is not None:
        anyscale.schedule.set_state(id=id, state=ScheduleState.DISABLED)
    else:
        if config_file is not None:
            name, cloud, project = _read_identifiers_from_config_file(config_file)

        anyscale.schedule.set_state(
            name=name, cloud=cloud, project=project, state=ScheduleState.DISABLED
        )


@schedule_cli.command(name="resume")
@click.option(
    "--config-file",
    "-f",
    required=False,
    type=str,
    help="Path to a YAML config file to use for this schedule.",
)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The named Anyscale Cloud for the schedule. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the schedule. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
def resume(
    config_file: str, name: str, cloud: str, project: str, id: str  # noqa: A002
) -> None:
    """ Resume a Schedule

    You can resume a schedule by config file, name, or id.

    To specify the schedule by name, use the --name flag. You can specify the cloud with --cloud and the project with --project.

    To specify the schedule by id, use the --id flag.

    To specify the schedule by config file, use --config-file. Ensure that name and optionally cloud and project are specified in the
    config file's job config.
    """
    _validate_schedule_identifiers(name=name, id=id, config_file=config_file)

    if id is not None:
        anyscale.schedule.set_state(id=id, state=ScheduleState.ENABLED)
    else:
        if config_file is not None:
            name, cloud, project = _read_identifiers_from_config_file(config_file)

        anyscale.schedule.set_state(
            name=name, cloud=cloud, project=project, state=ScheduleState.ENABLED
        )


@schedule_cli.command(name="status")
@click.option(
    "--config-file",
    "-f",
    required=False,
    type=str,
    help="Path to a YAML config file to use for this schedule.",
)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The named Anyscale Cloud for the schedule. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the schedule. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "--json",
    "-j",
    is_flag=True,
    default=False,
    help="Output the status in a structured JSON format.",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Include verbose details in the status.",
)
def status(
    config_file: str,
    name: str,
    cloud: str,
    project: str,
    id: str,  # noqa: A002
    json: bool,
    verbose: bool,
) -> None:
    """Query the status of a Schedule.

    You can query the status of a schedule by config file, name, or id.

    To specify the schedule by name, use the --name flag. You can specify the cloud with --cloud and the project with --project.

    To specify the schedule by id, use the --id flag.

    To specify the schedule by config file, use --config-file. Ensure that name and optionally cloud and project are specified in the
    config file's job config.
    """
    _validate_schedule_identifiers(name=name, id=id, config_file=config_file)

    if id is not None:
        status = anyscale.schedule.status(id=id)
    else:
        if config_file is not None:
            name, cloud, project = _read_identifiers_from_config_file(config_file)

        status = anyscale.schedule.status(name=name, cloud=cloud, project=project)

    status_dict = status.to_dict()

    if not verbose:
        status_dict.pop("config", None)

    if json:
        print(json_dumps(status_dict, indent=4, sort_keys=False))
    else:
        stream = StringIO()
        yaml.dump(status_dict, stream, sort_keys=False)
        print(stream.getvalue(), end="")


@schedule_cli.command(name="run",)
@click.option(
    "--config-file",
    "-f",
    required=False,
    type=str,
    help="Path to a YAML config file to use for this schedule.",
)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The named Anyscale Cloud for the schedule. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the schedule. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
def trigger(
    config_file: str, name: str, id: str, cloud: str, project: str  # noqa: A002
) -> None:
    """ Manually run a Schedule

    This function takes an existing schedule and runs it now.
    You can specify the schedule by name or id.
    You can also pass in a YAML file as a convinience. This is equivalent to passing in the name specified in the YAML file.
    IMPORTANT: if you pass in a YAML definition that differs from the Schedule defition, the Schedule will NOT be updated.
    Please use the `anyscale schedule apply` command to update the configuration of your schedule
    or use the `anyscale job submit` command to submit a one off job that is not a part of a schedule.
    """

    _validate_schedule_identifiers(name=name, id=id, config_file=config_file)

    if id is not None:
        anyscale.schedule.trigger(id=id)
    else:
        if config_file is not None:
            name, cloud, project = _read_identifiers_from_config_file(config_file)

        anyscale.schedule.trigger(
            name=name, cloud=cloud, project=project,
        )


@schedule_cli.command(name="url",)
@click.argument("schedule_config_file", required=False)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
def url(schedule_config_file: str, id: str, name: str) -> None:  # noqa: A002
    """ Get a Schedule URL

    This function accepts 1 argument, a path to a YAML config file that defines this schedule.
    You can also specify the schedule by name or id.
    """

    job_controller = ScheduleController()
    id = job_controller.resolve_file_name_or_id(  # noqa: A001
        schedule_config_file=schedule_config_file, id=id, name=name
    )
    job_controller.url(id)



import re
from typing import List, Optional

import click

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import ClusterManagementStackVersions
from anyscale.client.openapi_client.models.compute_stack import ComputeStack
from anyscale.commands import cloud_commands_util
from anyscale.commands.util import OptionPromptNull
from anyscale.controllers.cloud_controller import CloudController
from anyscale.util import validate_non_negative_arg


log = BlockLogger()  # CLI Logger


@click.group(
    "cloud",
    short_help="Configure cloud provider authentication for Anyscale.",
    help="""Configure cloud provider authentication and setup
to allow Anyscale to launch instances in your account.""",
)
def cloud_cli() -> None:
    pass


@cloud_cli.command(name="delete", help="Delete a cloud.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Delete cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to delete. Alternative to cloud name.",
    required=False,
)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Don't ask for confirmation."
)
def cloud_delete(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str], yes: bool
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().delete_cloud(
        cloud_name=cloud_name or name, cloud_id=cloud_id, skip_confirmation=yes
    )


@cloud_cli.command(
    name="set-default",
    help=(
        "Sets default cloud for your organization. This operation can only be performed "
        "by organization admins, and the default cloud must have organization level "
        "permissions."
    ),
)
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Set cloud as default by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to set as default. Alternative to cloud name.",
    required=False,
)
def cloud_set_default(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str]
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().set_default_cloud(
        cloud_name=cloud_name or name, cloud_id=cloud_id
    )


@cloud_cli.command(name="setup", help="Set up a cloud provider.")
@click.option(
    "--provider",
    help="The cloud provider type.",
    required=True,
    prompt="Provider",
    type=click.Choice(["aws", "gcp"], case_sensitive=False),
)
@click.option(
    "--region",
    cls=OptionPromptNull,
    help="Region to set up the credentials in.",
    required=True,
    prompt="Region",
    default_option="provider",
    default=lambda p: "us-west-2" if p == "aws" else "us-west1",
    show_default=True,
)
@click.option("--name", "-n", help="Name of the cloud.", required=True, prompt="Name")
@click.option(
    "--project-id",
    help="Globally Unique project ID for GCP clouds (e.g., my-project-abc123)",
    required=False,
    type=str,
)
@click.option(
    "--functional-verify",
    help="Verify the cloud is functional. This will check that the cloud can launch workspace/service.",
    required=False,
    is_flag=False,
    flag_value="workspace",
)
@click.option(
    "--anyscale-managed",
    is_flag=True,
    default=False,
    help="Let anyscale create all the resources.",
)
@click.option(
    "--enable-head-node-fault-tolerance",
    is_flag=True,
    default=False,
    help="Whether to enable head node fault tolerance for services.",
)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Skip asking for confirmation."
)
@click.option(
    "--disable-auto-add-user",
    is_flag=True,
    default=False,
    help=(
        "All users in the organization will be added to clouds created "
        "with `anyscale cloud setup` by default. Specify --disable-auto-add-user to "
        "disable this and instead manually grant users permissions to the cloud."
    ),
)
def setup_cloud(  # noqa: PLR0913
    provider: str,
    region: str,
    name: str,
    project_id: str,
    functional_verify: Optional[str],
    anyscale_managed: bool,  # noqa: ARG001
    enable_head_node_fault_tolerance: bool,
    yes: bool,
    disable_auto_add_user: bool,
) -> None:
    # TODO (congding): remove `anyscale_managed` in the future, now keeping it for compatibility
    if provider == "aws":
        CloudController().setup_managed_cloud(
            provider=provider,
            region=region,
            name=name,
            functional_verify=functional_verify,
            cluster_management_stack_version=ClusterManagementStackVersions.V2,
            enable_head_node_fault_tolerance=enable_head_node_fault_tolerance,
            yes=yes,
            auto_add_user=(not disable_auto_add_user),
        )
    elif provider == "gcp":
        if not project_id:
            project_id = click.prompt("GCP Project ID", type=str)
        if project_id[0].isdigit():
            # project ID should start with a letter
            raise click.ClickException(
                "Please provide a valid project ID. Note that project ID is not project number, see https://cloud.google.com/resource-manager/docs/creating-managing-projects#before_you_begin for details."
            )
        CloudController().setup_managed_cloud(
            provider=provider,
            region=region,
            name=name,
            project_id=project_id,
            functional_verify=functional_verify,
            cluster_management_stack_version=ClusterManagementStackVersions.V2,
            enable_head_node_fault_tolerance=enable_head_node_fault_tolerance,
            yes=yes,
            auto_add_user=(not disable_auto_add_user),
        )


@cloud_cli.command(
    name="list", help=("List information about clouds in your Anyscale organization."),
)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="Name of cloud to get information about.",
)
@click.option(
    "--cloud-id",
    "--id",
    required=False,
    default=None,
    help=("Id of cloud to get information about."),
)
@click.option(
    "--max-items",
    required=False,
    default=20,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list_cloud(name: Optional[str], cloud_id: Optional[str], max_items: int,) -> None:
    print(
        CloudController().list_clouds(
            cloud_name=name, cloud_id=cloud_id, max_items=max_items
        )
    )


@cloud_cli.group("config", help="Manage the configuration for a cloud.")
def cloud_config_group() -> None:
    pass


@cloud_cli.command(
    name="update",
    help=(
        "Update a managed cloud to the latest configuration. Only applicable for anyscale managed clouds."
    ),
)
@click.argument("cloud-name", required=False)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to update. Alternative to cloud name.",
    required=False,
)
@click.option("--name", "-n", help="Update configuration of cloud by name.", type=str)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Skip asking for confirmation."
)
@click.option(
    "--functional-verify",
    help="Verify the cloud is functional. This will check that the cloud can launch workspace/service.",
    required=False,
    is_flag=False,
    flag_value="workspace",
)
@click.option(
    "--enable-head-node-fault-tolerance",
    is_flag=True,
    default=False,
    help="Whether to enable head node fault tolerance for services.",
)
@click.option(
    "--enable-auto-add-user/--disable-auto-add-user",
    default=None,
    help=(
        "If --enable-auto-add-user is specified for a cloud, all users in the organization "
        "will be added to the cloud by default. Note: There may be up to 30 sec delay for all users to be granted "
        "permissions after this feature is enabled.\n\n"
        "Specifying --disable-auto-add-user will require that users "
        "are manually granted permissions to access the cloud. No existing cloud permissions are altered by specifying this flag."
    ),
)
def cloud_update(
    cloud_name: Optional[str],
    name: Optional[str],
    cloud_id: Optional[str],
    functional_verify: Optional[str],
    enable_head_node_fault_tolerance: bool,
    yes: bool,
    enable_auto_add_user: Optional[bool],
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    if enable_head_node_fault_tolerance and (enable_auto_add_user is not None):
        raise click.ClickException(
            "Please only specify either --enable-head-node-fault-tolerance or "
            f"{'--enable-auto-add-user' if enable_auto_add_user else '--disable-auto-add-user'} for "
            "this call of `anyscale cloud update`. The other flag can be specified in a separate call "
            "to the command."
        )
    CloudController().update_managed_cloud(
        cloud_name=cloud_name or name,
        cloud_id=cloud_id,
        enable_head_node_fault_tolerance=enable_head_node_fault_tolerance,
        functional_verify=functional_verify,
        yes=yes,
        auto_add_user=enable_auto_add_user,
    )


@cloud_config_group.command("get", help="Get the current configuration for a cloud.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Update configuration of cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to get details about. Alternative to cloud name.",
    required=False,
)
def cloud_config_get(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str]
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    print(
        CloudController().get_cloud_config(
            cloud_name=cloud_name or name, cloud_id=cloud_id,
        )
    )


@cloud_config_group.command(
    "update", help="Update the current configuration for a cloud."
)
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Update configuration of cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to get details about. Alternative to cloud name.",
    required=False,
)
@click.option(
    "--enable-log-ingestion/--disable-log-ingestion",
    default=None,
    help=(
        "If --enable-log-ingestion is specified for a cloud, it will enable the log "
        "viewing and querying UI features for the clusters on this cloud. This will "
        "enable easier debugging. The logs produced by the clusters will "
        "be sent from the data plane to the control plane. Anyscale does not share "
        "this data with any third party or use it for any purpose other than serving "
        "the log UI for the customer. The log will be stored at most 30 days."
        "Please note by disable this feature again, Anyscale doesn't "
        "delete the logs that have already been ingested. Your clusters may incur "
        "extra data transfer cost from the cloud provider by enabling this feature."
    ),
)
def cloud_config_update(
    cloud_name: Optional[str],
    name: Optional[str],
    cloud_id: Optional[str],
    enable_log_ingestion: Optional[bool],
) -> None:
    if any([enable_log_ingestion is not None]):
        if enable_log_ingestion is True:
            consent_message = click.prompt(
                "--enable-log-ingestion is specified. Please note the logs produced by "
                "your cluster will be ingested into Anyscale's service in region "
                "us-west-2. Your clusters may incur extra data transfer cost from the "
                "cloud provider. If you are sure you want to enable this feature, "
                'please type "consent"',
                type=str,
            )
            if consent_message != "consent":
                raise click.ClickException(
                    'You must type "consent" to enable log ingestion.'
                )
        if enable_log_ingestion is False:
            confirm_response = click.confirm(
                "--disable-log-ingestion is specified. Please note the logs that's "
                "already ingested will not be deleted. Existing clusters will not stop"
                "the log ingestion until you restart them. Logs are automatically "
                "deleted after 30 days from the time of ingestion. Are you sure you "
                "want to disable log ingestion?"
            )
            if not confirm_response:
                raise click.ClickException("You must confirm to disable log ingestion.")

        CloudController().update_cloud_config(
            cloud_name=cloud_name or name,
            cloud_id=cloud_id,
            enable_log_ingestion=enable_log_ingestion,
        )
    else:
        raise click.ClickException(
            "Please provide at least one of the following arguments: --enable-log-ingestion, --disable-log-ingestion."
        )


@cloud_cli.command(
    name="register", help="Register an anyscale cloud with your own resources."
)
@click.option(
    "--provider",
    help="The cloud provider type.",
    required=True,
    type=click.Choice(["aws", "gcp"], case_sensitive=False),
)
@click.option(
    "--region",
    cls=OptionPromptNull,
    help="Region to set up the credentials in.",
    required=True,
    default_option="provider",
    default=lambda p: "us-west-2" if p == "aws" else "us-west1",
    show_default=True,
)
@click.option(
    "--compute-stack",
    help="The compute stack type (VM or K8S).",
    required=False,
    type=click.Choice([ComputeStack.VM, ComputeStack.K8S], case_sensitive=False),
    default=ComputeStack.VM,
    # TODO (shomilj): Unhide this option when full support for Kubernetes has been rolled out.
    hidden=True,
)
@click.option(
    "--name", "-n", help="Name of the cloud.", required=True,
)
@click.option(
    "--vpc-id", help="The ID of the VPC.", required=False, type=str,
)
@click.option(
    "--subnet-ids",
    help="Comma separated list of subnet ids.",
    required=False,
    type=str,
)
@click.option(
    "--efs-id", help="The EFS ID.", required=False, type=str,
)
@click.option(
    "--anyscale-iam-role-id",
    help="The Anyscale IAM Role ARN.",
    required=False,
    type=str,
)
@click.option(
    "--instance-iam-role-id",
    help="The instance IAM role ARN.",
    required=False,
    type=str,
)
@click.option(
    "--security-group-ids",
    help="IDs of the security groups.",
    required=False,
    type=str,
)
@click.option(
    "--s3-bucket-id", help="S3 bucket ID.", required=False, type=str,
)
@click.option(
    "--external-id",
    help="The trust policy external ID for the cross account IAM role.",
    required=False,
    type=str,
)
@click.option(
    "--memorydb-cluster-id", help="Memorydb cluster ID", required=False, type=str,
)
@click.option(
    "--project-id",
    help="Globally Unique project ID for GCP clouds (e.g., my-project-abc123)",
    required=False,
    type=str,
)
@click.option(
    "--vpc-name", help="VPC name for GCP clouds", required=False, type=str,
)
@click.option(
    "--subnet-names",
    help="Comma separated list of subnet names for GCP clouds",
    required=False,
    type=str,
)
@click.option(
    "--filestore-instance-id",
    help="Filestore instance ID for GCP clouds.",
    required=False,
    type=str,
)
@click.option(
    "--filestore-location",
    help="Filestore location for GCP clouds.",
    required=False,
    type=str,
)
@click.option(
    "--anyscale-service-account-email",
    help="Anyscale service account email for GCP clouds.",
    required=False,
    type=str,
)
@click.option(
    "--instance-service-account-email",
    help="Instance service account email for GCP clouds.",
    required=False,
    type=str,
)
@click.option(
    "--provider-name",
    help="Workload Identity Federation provider name for Anyscale access.",
    required=False,
    type=str,
)
@click.option(
    "--firewall-policy-names",
    help="Filewall policy names for GCP clouds",
    required=False,
    type=str,
)
@click.option(
    "--cloud-storage-bucket-name",
    help="Cloud storage bucket name for GCP clouds",
    required=False,
    type=str,
)
@click.option(
    "--memorystore-instance-name",
    help="Memorystore instance name for GCP clouds",
    required=False,
    type=str,
)
@click.option(
    "--host-project-id",
    help="Host project ID for shared VPC",
    required=False,
    type=str,
)
@click.option(
    "--kubernetes-namespaces",
    help="On the Kubernetes compute stack, a comma-separated list of namespace(s) that can be used for running workloads. The first namespace in the list will be used as the default.",
    required=False,
    type=str,
)
@click.option(
    "--kubernetes-ingress-external-address",
    help="On the Kubernetes compute stack, the address of the ingress used for serving traffic for Anyscale services.",
    required=False,
    type=str,
)
@click.option(
    "--kubernetes-zones",
    help="On the Kubernetes compute stack, a comma-separated list of zones to launch pods in.",
    required=False,
    type=str,
)
@click.option(
    "--kubernetes-dataplane-identity",
    help="On the Kubernetes compute stack, the IAM identity for the Kubernetes Manager deployment. IN AWS EKS, this is the ARN of the IAM role associated with the Kubernetes Manager deployment. For GCP GKE, this is the service account email associated with the Kubernetes Manager deployment.",
    required=False,
    type=str,
)
@click.option(
    "--private-network", help="Use private network.", is_flag=True, default=False,
)
@click.option(
    "--functional-verify",
    help="Verify the cloud is functional. This will check that the cloud can launch workspace/service.",
    required=False,
    is_flag=False,
    flag_value="workspace",
)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Skip asking for confirmation."
)
@click.option(
    "--skip-verifications",
    help="Skip verifications. This will skip all verifications.",
    required=False,
    is_flag=True,
    type=bool,
    default=False,
)
@click.option(
    "--enable-auto-add-user",
    is_flag=True,
    default=False,
    help=(
        "If --enable-auto-add-user is specified for a cloud, all users in the organization "
        "will be added to the cloud by default. Otherwise users will need to be manually granted "
        "permissions to the cloud. Note: There may be up to 30 sec delay for all users to be granted "
        "permissions after the cloud is created."
    ),
)
def register_cloud(  # noqa: PLR0913, PLR0912
    provider: str,
    region: str,
    compute_stack: ComputeStack,
    name: str,
    vpc_id: str,
    subnet_ids: str,
    efs_id: str,
    anyscale_iam_role_id: str,
    instance_iam_role_id: str,
    security_group_ids: str,
    s3_bucket_id: str,
    external_id: Optional[str],
    memorydb_cluster_id: str,
    project_id: str,
    vpc_name: str,
    subnet_names: str,
    filestore_instance_id: str,
    filestore_location: str,
    anyscale_service_account_email: str,
    instance_service_account_email: str,
    provider_name: str,
    firewall_policy_names: str,
    cloud_storage_bucket_name: str,
    memorystore_instance_name: str,
    host_project_id: Optional[str],
    kubernetes_namespaces: Optional[str],
    kubernetes_ingress_external_address: Optional[str],
    kubernetes_zones: Optional[str],
    kubernetes_dataplane_identity: Optional[str],
    functional_verify: Optional[str],
    private_network: bool,
    yes: bool,
    skip_verifications: bool,
    enable_auto_add_user: bool,
) -> None:
    missing_args: List[str] = []
    if provider == "aws":
        # Check for missing required arguments for AWS clouds,
        # based on the compute stack (not all args are required
        # on all compute stacks).
        for resource in [
            (vpc_id, "--vpc-id", (ComputeStack.VM)),
            (subnet_ids, "--subnet-ids", (ComputeStack.VM)),
            (efs_id, "--efs-id", (ComputeStack.VM)),
            (
                anyscale_iam_role_id,
                "--anyscale-iam-role-id",
                (ComputeStack.VM, ComputeStack.K8S),
            ),
            (instance_iam_role_id, "--instance-iam-role-id", (ComputeStack.VM)),
            (security_group_ids, "--security-group-ids", (ComputeStack.VM)),
            (s3_bucket_id, "--s3-bucket-id", (ComputeStack.VM, ComputeStack.K8S)),
            (kubernetes_namespaces, "--kubernetes-namespaces", (ComputeStack.K8S)),
            (
                kubernetes_ingress_external_address,
                "--kubernetes-ingress-external-address",
                (ComputeStack.K8S),
            ),
            (kubernetes_zones, "--kubernetes-zones", (ComputeStack.K8S)),
            (
                kubernetes_dataplane_identity,
                "--kubernetes-dataplane-identity",
                (ComputeStack.K8S),
            ),
        ]:
            if compute_stack in resource[2] and resource[0] is None:
                missing_args.append(resource[1])

        if len(missing_args) > 0:
            raise click.ClickException(f"Please provide a value for {missing_args}")

        CloudController().register_aws_cloud(
            region=region,
            compute_stack=compute_stack,
            name=name,
            vpc_id=vpc_id,
            subnet_ids=subnet_ids.split(",") if subnet_ids else [],
            efs_id=efs_id,
            anyscale_iam_role_id=anyscale_iam_role_id,
            instance_iam_role_id=instance_iam_role_id,
            security_group_ids=security_group_ids.split(",")
            if security_group_ids
            else [],
            s3_bucket_id=s3_bucket_id,
            memorydb_cluster_id=memorydb_cluster_id,
            functional_verify=functional_verify,
            kubernetes_namespaces=kubernetes_namespaces.split(",")
            if kubernetes_namespaces
            else [],
            kubernetes_ingress_external_address=kubernetes_ingress_external_address,
            kubernetes_zones=kubernetes_zones.split(",") if kubernetes_zones else [],
            kubernetes_dataplane_identity=kubernetes_dataplane_identity,
            private_network=private_network,
            cluster_management_stack_version=ClusterManagementStackVersions.V2,
            yes=yes,
            skip_verifications=skip_verifications,
            auto_add_user=enable_auto_add_user,
            external_id=external_id,
        )
    elif provider == "gcp":
        # Keep the parameter naming ({resource}_name or {resource}_id) consistent with GCP to reduce confusion for customers
        # Check if all required parameters are provided
        # memorystore_instance_name and host_project_id are optional for GCP clouds
        for resource in [
            (project_id, "--project-id", (ComputeStack.VM, ComputeStack.K8S)),
            (vpc_name, "--vpc-name", (ComputeStack.VM)),
            (subnet_names, "--subnet-names", (ComputeStack.VM)),
            (filestore_instance_id, "--filestore-instance-id", (ComputeStack.VM)),
            (filestore_location, "--filestore-location", (ComputeStack.VM)),
            (
                anyscale_service_account_email,
                "--anyscale-service-account-email",
                (ComputeStack.VM, ComputeStack.K8S),
            ),
            (
                instance_service_account_email,
                "--instance-service-account-email",
                (ComputeStack.VM),
            ),
            (provider_name, "--provider-name", (ComputeStack.VM, ComputeStack.K8S)),
            (firewall_policy_names, "--firewall-policy-names", (ComputeStack.VM)),
            (
                cloud_storage_bucket_name,
                "--cloud-storage-bucket-name",
                (ComputeStack.VM, ComputeStack.K8S),
            ),
            (kubernetes_namespaces, "--kubernetes-namespaces", (ComputeStack.K8S)),
            (
                kubernetes_ingress_external_address,
                "--kubernetes-ingress-external-address",
                (ComputeStack.K8S),
            ),
            (kubernetes_zones, "--kubernetes-zones", (ComputeStack.K8S)),
            (
                kubernetes_dataplane_identity,
                "--kubernetes-dataplane-identity",
                (ComputeStack.K8S),
            ),
        ]:
            if compute_stack in resource[2] and resource[0] is None:
                missing_args.append(resource[1])

        if len(missing_args) > 0:
            raise click.ClickException(f"Please provide a value for {missing_args}")

        if project_id[0].isdigit():
            # project ID should start with a letter
            raise click.ClickException(
                "Please provide a valid project ID. Note that project ID is not project number, see https://cloud.google.com/resource-manager/docs/creating-managing-projects#before_you_begin for details."
            )

        if (
            re.search(
                "projects/[0-9]*/locations/global/workloadIdentityPools/.+/providers/.+",
                provider_name,
            )
            is None
        ):
            raise click.ClickException(
                "Please provide a valid, fully qualified provider name. Example: projects/<project number>/locations/global/workloadIdentityPools/<pool name>/providers/<provider id>"
            )

        if (
            memorystore_instance_name is not None
            and not cloud_commands_util.validate_memorystore_instance_name(
                memorystore_instance_name
            )
        ):
            raise click.ClickException(
                "Please provide a valid memorystore instance name. Example: projects/<project number>/locations/<location>/instances/<instance id>"
            )

        if host_project_id is not None and host_project_id[0].isdigit():
            # project ID should start with a letter
            raise click.ClickException(
                "Please provide a valid project ID for `--host-project-id`. Note that project ID is not project number, see https://cloud.google.com/resource-manager/docs/creating-managing-projects#before_you_begin for details."
            )

        CloudController().register_gcp_cloud(
            region=region,
            name=name,
            compute_stack=compute_stack,
            project_id=project_id,
            vpc_name=vpc_name,
            subnet_names=subnet_names.split(",") if subnet_names else [],
            filestore_instance_id=filestore_instance_id,
            filestore_location=filestore_location,
            anyscale_service_account_email=anyscale_service_account_email,
            instance_service_account_email=instance_service_account_email,
            # TODO (allenyin): use provider_name instead of provider_id everywhere.
            provider_id=provider_name,
            firewall_policy_names=firewall_policy_names.split(",")
            if firewall_policy_names
            else [],
            cloud_storage_bucket_name=cloud_storage_bucket_name,
            memorystore_instance_name=memorystore_instance_name,
            host_project_id=host_project_id,
            kubernetes_namespaces=kubernetes_namespaces.split(",")
            if kubernetes_namespaces
            else [],
            kubernetes_ingress_external_address=kubernetes_ingress_external_address,
            kubernetes_zones=kubernetes_zones.split(",") if kubernetes_zones else [],
            kubernetes_dataplane_identity=kubernetes_dataplane_identity,
            functional_verify=functional_verify,
            private_network=private_network,
            cluster_management_stack_version=ClusterManagementStackVersions.V2,
            yes=yes,
            skip_verifications=skip_verifications,
            auto_add_user=enable_auto_add_user,
        )
    else:
        raise click.ClickException(
            f"Invalid Cloud provider: {provider}. Available providers are [aws, gcp]."
        )


@cloud_cli.command(name="verify", help="Checks the healthiness of a cloud.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Verify cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Verify cloud by cloud id, alternative to cloud name.",
    required=False,
)
@click.option(
    "--functional-verify",
    help="Verify the cloud is functional. This will check that the cloud can launch workspace/service.",
    required=False,
    is_flag=False,
    flag_value="workspace",
)
@click.option(
    "--strict",
    is_flag=True,
    default=False,
    help="Strict Verify. Treat warnings as failures.",
)
def cloud_verify(
    cloud_name: Optional[str],
    name: Optional[str],
    cloud_id: Optional[str],
    functional_verify: Optional[str],
    strict: bool = False,
) -> bool:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )

    return CloudController().verify_cloud(
        cloud_name=cloud_name or name,
        cloud_id=cloud_id,
        functional_verify=functional_verify,
        strict=strict,
    )


@cloud_cli.command(
    name="edit",
    help="Edit registered cloud resource on Anyscale. Only applicable for anyscale registered clouds.",
)
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Edit cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Edit cloud by id, alternative to cloud name.",
    required=False,
)
@click.option(
    "--aws-s3-id", help="New S3 bucket ID.", required=False, type=str,
)
@click.option("--aws-efs-id", help="New EFS ID.", required=False, type=str)
@click.option(
    "--aws-efs-mount-target-ip",
    help="New EFS mount target IP.",
    required=False,
    type=str,
)
@click.option(
    "--memorydb-cluster-id",
    help="New AWS Memorydb cluster ID.",
    required=False,
    type=str,
)
@click.option(
    "--gcp-filestore-instance-id",
    help="New GCP filestore instance id.",
    required=False,
    type=str,
)
@click.option(
    "--gcp-filestore-location",
    help="New GCP filestore location.",
    required=False,
    type=str,
)
@click.option(
    "--gcp-cloud-storage-bucket-name",
    help="New GCP Cloud storage bucket name.",
    required=False,
    type=str,
)
@click.option(
    "--memorystore-instance-name",
    help="New Memorystore instance name for GCP clouds",
    required=False,
    type=str,
)
@click.option(
    "--functional-verify",
    help="Verify the cloud is functional. This will check that the cloud can launch workspace/service.",
    required=False,
    is_flag=False,
    flag_value="workspace",
)
@click.option(
    "--enable-auto-add-user/--disable-auto-add-user",
    default=None,
    help=(
        "If --enable-auto-add-user is specified for a cloud, all users in the organization "
        "will be added to the cloud by default. Note: There may be up to 30 sec delay for all users to be granted "
        "permissions after this feature is enabled.\n\n"
        "Specifying --disable-auto-add-user will require that users "
        "are manually granted permissions to access the cloud. No existing cloud permissions are altered by specifying this flag."
    ),
)
def cloud_edit(  # noqa: PLR0913
    cloud_name: Optional[str],
    name: Optional[str],
    cloud_id: Optional[str],
    aws_s3_id: Optional[str],
    aws_efs_id: Optional[str],
    aws_efs_mount_target_ip: Optional[str],
    memorydb_cluster_id: Optional[str],
    gcp_filestore_instance_id: Optional[str],
    gcp_filestore_location: Optional[str],
    gcp_cloud_storage_bucket_name: Optional[str],
    memorystore_instance_name: Optional[str],
    functional_verify: Optional[str],
    enable_auto_add_user: Optional[bool],
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    if any(
        [
            aws_s3_id,
            aws_efs_id,
            aws_efs_mount_target_ip,
            memorydb_cluster_id,
            gcp_filestore_instance_id,
            gcp_filestore_location,
            gcp_cloud_storage_bucket_name,
            memorystore_instance_name,
            enable_auto_add_user is not None,
        ]
    ):
        if any([gcp_filestore_instance_id, gcp_filestore_location]) and not all(
            [gcp_filestore_instance_id, gcp_filestore_location]
        ):
            # Make sure both gcp_filestore_instance_id and gcp_filestore_location are provided if you want to edit filestore.
            raise click.ClickException(
                "Please provide both --gcp-filestore-instance-id and --gcp-filestore-location if you want to edit filestore."
            )
        if (
            memorystore_instance_name is not None
            and not cloud_commands_util.validate_memorystore_instance_name(
                memorystore_instance_name
            )
        ):
            raise click.ClickException(
                "Please provide a valid memorystore instance name. Example: projects/<project number>/locations/<location>/instances/<instance id>"
            )
        CloudController().edit_cloud(
            cloud_name=cloud_name or name,
            cloud_id=cloud_id,
            aws_s3_id=aws_s3_id,
            aws_efs_id=aws_efs_id,
            aws_efs_mount_target_ip=aws_efs_mount_target_ip,
            memorydb_cluster_id=memorydb_cluster_id,
            gcp_filestore_instance_id=gcp_filestore_instance_id,
            gcp_filestore_location=gcp_filestore_location,
            gcp_cloud_storage_bucket_name=gcp_cloud_storage_bucket_name,
            memorystore_instance_name=memorystore_instance_name,
            functional_verify=functional_verify,
            auto_add_user=enable_auto_add_user,
        )
    else:
        raise click.ClickException(
            "Please provide at least one of the following arguments: --aws-s3-id, --aws-efs-id, --aws-efs-mount-target-ip, --memorydb-cluster-id, --gcp-filestore-instance-id, --gcp-filestore-location, --gcp-cloud-storage-bucket-name, --memorystore-instance-name, --enable-auto-add-user, --disable-auto-add-user."
        )



from typing import Tuple

from common import OPENAPI_NO_VALIDATION, TestLogger
import pytest

from anyscale import Anyscale
from anyscale._private.anyscale_client import FakeAnyscaleClient
from anyscale._private.sdk.timer import FakeTimer
from anyscale.client.openapi_client.models.decorated_schedule import DecoratedSchedule
from anyscale.client.openapi_client.models.production_job_config import (
    ProductionJobConfig,
)
from anyscale.client.openapi_client.models.schedule_config import (
    ScheduleConfig as BackendScheduleConfig,
)
from anyscale.job.models import JobConfig
from anyscale.schedule import ScheduleSDK
from anyscale.schedule.models import ScheduleConfig, ScheduleState, ScheduleStatus


@pytest.fixture()
def schedule_sdk_with_fakes(
    sdk_with_fakes: Tuple[Anyscale, FakeAnyscaleClient, TestLogger, FakeTimer]
) -> Tuple[ScheduleSDK, FakeAnyscaleClient, TestLogger, FakeTimer]:
    sdk, client, logger, timer = sdk_with_fakes
    return sdk.schedule, client, logger, timer


class TestApply:
    def test_basic(
        self,
        schedule_sdk_with_fakes: Tuple[
            ScheduleSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = schedule_sdk_with_fakes

        job_config = JobConfig(entrypoint="python hello.py", name="test-job-name",)
        config = ScheduleConfig(
            job_config=job_config, cron_expression="0 0 * * * *", timezone="UTC",
        )

        schedule_id = sdk.apply(config)
        created_schedule = fake_client.get_schedule(
            id=schedule_id, name=None, cloud=None, project=None
        )

        expected_schedule = DecoratedSchedule(
            id=schedule_id,
            name=job_config.name,
            project_id=fake_client.DEFAULT_PROJECT_ID,
            config=ProductionJobConfig(
                entrypoint=job_config.entrypoint,
                build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
                max_retries=1,
                compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
            schedule=BackendScheduleConfig(
                cron_expression=config.cron_expression, timezone=config.timezone
            ),
            next_trigger_at=fake_client.SCHEDULE_NEXT_TRIGGER_AT_TIME,
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )

        assert created_schedule is not None
        assert created_schedule == expected_schedule


class TestSetState:
    def test_basic(
        self,
        schedule_sdk_with_fakes: Tuple[
            ScheduleSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = schedule_sdk_with_fakes

        name = "test-job-name"
        job_config = JobConfig(entrypoint="python hello.py", name=name)
        config = ScheduleConfig(
            job_config=job_config, cron_expression="0 0 * * * *", timezone="UTC",
        )

        schedule_id = sdk.apply(config)
        assert fake_client.schedule_is_enabled(schedule_id)

        sdk.set_state(id=schedule_id, state=ScheduleState.DISABLED)
        assert not fake_client.schedule_is_enabled(schedule_id)

        sdk.set_state(id=schedule_id, state=ScheduleState.ENABLED)
        assert fake_client.schedule_is_enabled(schedule_id)

        sdk.set_state(name=name, state=ScheduleState.DISABLED)
        assert not fake_client.schedule_is_enabled(schedule_id)

        sdk.set_state(name=name, state=ScheduleState.ENABLED)
        assert fake_client.schedule_is_enabled(schedule_id)


class TestStatus:
    def test_schedule_not_found(
        self,
        schedule_sdk_with_fakes: Tuple[
            ScheduleSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = schedule_sdk_with_fakes

        with pytest.raises(
            RuntimeError, match="Schedule with name 'test-schedule-name' was not found."
        ):
            sdk.status(name="test-schedule-name")

    def test_basic(
        self,
        schedule_sdk_with_fakes: Tuple[
            ScheduleSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = schedule_sdk_with_fakes

        name = "test-job-name"
        job_config = JobConfig(
            entrypoint="python hello.py",
            name=name,
            compute_config=f"{fake_client.DEFAULT_CLUSTER_COMPUTE_NAME}:1",
        )
        config = ScheduleConfig(
            job_config=job_config, cron_expression="0 0 * * * *", timezone="UTC",
        )

        schedule_id = sdk.apply(config)

        expected_schedule_status = ScheduleStatus(
            config=config, name=name, id=schedule_id, state=ScheduleState.ENABLED
        )
        assert sdk.status(id=schedule_id) == expected_schedule_status

        sdk.set_state(id=schedule_id, state=ScheduleState.DISABLED)

        expected_schedule_status = ScheduleStatus(
            config=config, name=name, id=schedule_id, state=ScheduleState.DISABLED
        )
        assert sdk.status(id=schedule_id) == expected_schedule_status


class TestTrigger:
    def test_basic(
        self,
        schedule_sdk_with_fakes: Tuple[
            ScheduleSDK, FakeAnyscaleClient, TestLogger, FakeTimer
        ],
    ):
        sdk, fake_client, _, _ = schedule_sdk_with_fakes

        name = "test-job-name"
        job_config = JobConfig(
            entrypoint="python hello.py",
            name=name,
            compute_config=f"{fake_client.DEFAULT_CLUSTER_COMPUTE_NAME}:1",
        )
        config = ScheduleConfig(
            job_config=job_config, cron_expression="0 0 * * * *", timezone="UTC",
        )

        schedule_id = sdk.apply(config)
        assert fake_client.trigger_counts(schedule_id) == 0

        sdk.trigger(id=schedule_id)
        assert fake_client.trigger_counts(schedule_id) == 1

        sdk.trigger(name=name)
        assert fake_client.trigger_counts(schedule_id) == 2



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

from anyscale._private.models import ModelBase, ModelEnum
from anyscale._private.workload import WorkloadConfig


@dataclass(frozen=True)
class RayGCSExternalStorageConfig(ModelBase):
    """Configuration options for external storage for the Ray Global Control Store (GCS).

    When external storage is enabled for a service, it will be able to continue serving traffic if the
    head node goes down. Without external storage, the entire cluster will be restarted.
    """

    __doc_py_example__ = """\
from anyscale.service.models import ServiceConfig, RayGCSExternalStorageConfig

config = ServiceConfig(
    name="my-service",
    applications=[{"import_path": "main:app"}],
    ray_gcs_external_storage_config=RayGCSExternalStorageConfig(
        enabled=True,
        address="http://redis-address:8888",
        certificate_path="/etc/ssl/certs/ca-certificates.crt",
    ),
)
"""

    __doc_yaml_example__ = """\
name: my-service
applications:
  - import_path: main:app
ray_gcs_external_storage_config:
  enabled: true
  address: http://redis-address:8888
  certificate_path: /etc/ssl/certs/ca-certificates.crt
"""

    enabled: bool = field(
        default=True,
        metadata={
            "docstring": "Enable or disable external storage. When `False`, the other fields are ignored."
        },
    )

    def _validate_enabled(self, enabled: bool):
        if not isinstance(enabled, bool):
            raise TypeError("'enabled' must be a boolean.")

    address: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "Address to connect to the redis instance at. Defaults to the cloud-wide configuration."
        },
    )

    def _validate_address(self, address: Optional[str]):
        if address is not None and not isinstance(address, str):
            raise TypeError("'address' must be a string.")

    certificate_path: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "Path to the TLS certificate file to use for authentication when using secure connections. Defaults to the cloud-wide configuration."
        },
    )

    def _validate_certificate_path(self, certificate_path: Optional[str]):
        if certificate_path is not None and not isinstance(certificate_path, str):
            raise TypeError("'certificate_path' must be a string.")


@dataclass(frozen=True)
class TracingConfig(ModelBase):
    """Configuration options for tracing.

    Tracing provides a way to collect telemetry on requests that are
    handled by your Serve application. If tracing is enabled, you can instrument
    your Serve application and export traces to a specific backend by configuring the
    `exporter_import_path` option. This feature is currently experimental.
    """

    __doc_py_example__ = """\
from anyscale.service.models import ServiceConfig, TracingConfig

config = ServiceConfig(
    name="my-service",
    applications=[{"import_path": "main:app"}],
    tracing_config=TracingConfig(
        enabled=True,
        exporter_import_path="my_module:custom_tracing_exporter",
        sampling_ratio=1.0,
    ),
)
"""

    __doc_yaml_example__ = """\
name: my-service
applications:
  - import_path: main:app
tracing_config:
  enabled: true
  exporter_import_path: "my_module:custom_tracing_exporter"
  sampling_ratio: 1.0
"""

    enabled: bool = field(
        default=True,
        metadata={
            "docstring": (
                "Flag to enable tracing. "
                "When `True` and no `exporter_import_path` "
                "is defined, then the default tracing exporter will be "
                "used to export traces to serve logs. "
                "When `False`, the other fields are ignored."
            )
        },
    )

    def _validate_enabled(self, enabled: bool):
        if not isinstance(enabled, bool):
            raise TypeError("'enabled' must be a boolean.")

    exporter_import_path: Optional[str] = field(
        default=None,
        metadata={
            "docstring": (
                "Path to exporter function. Should be of the "
                'form "module.submodule_1...submodule_n.'
                'export_tracing". This is equivalent to '
                '"from module.submodule_1...submodule_n import export_tracing. '
                "Exporter function takes no arguments and returns List[SpanProcessor]."
            )
        },
    )

    def _validate_exporter_import_path(self, exporter_import_path: Optional[str]):
        if exporter_import_path is not None and not isinstance(
            exporter_import_path, str
        ):
            raise TypeError("'exporter_import_path' must be a string.")

    sampling_ratio: float = field(
        default=1.0,
        metadata={
            "docstring": "Ratio of traces to record and export. "
            "Reducing the ratio can help with minimizing serving latency and storage overhead. "
        },
    )

    def _validate_sampling_ratio(self, sampling_ratio: float):
        if sampling_ratio is not None and not isinstance(sampling_ratio, float):
            raise TypeError("'sampling_ratio' must be a float.")


@dataclass(frozen=True)
class ServiceConfig(WorkloadConfig):
    """Configuration options for a service."""

    __doc_py_example__ = """\
from anyscale.service.models import ServiceConfig

config = ServiceConfig(
    name="my-service",
    working_dir=".",
    applications=[{"import_path": "main:app"}],
    # An inline `ComputeConfig` can also be provided.
    compute_config="my-compute-config:1",
    # A containerfile path can also be provided.
    image_uri="anyscale/image/my-image:1",
)
"""

    __doc_yaml_example__ = """\
name: my-service
applications:
  - import_path: main:app
image_uri: anyscale/image/my-image:1 # (Optional) Exclusive with `containerfile`.
containerfile: /path/to/Dockerfile # (Optional) Exclusive with `image_uri`.
compute_config: my-compute-config:1 # (Optional) An inline dictionary can also be provided.
working_dir: /path/to/working_dir # (Optional) If a local directory is provided, it will be uploaded automatically.
excludes: # (Optional) List of files to exclude from being packaged up for the job.
  - .git
  - .env
  - .DS_Store
  - __pycache__
requirements: # (Optional) List of requirements files to install. Can also be a path to a requirements.txt.
  - emoji==1.2.0
  - numpy==1.19.5
env_vars: # (Optional) Dictionary of environment variables to set in the job.
  MY_ENV_VAR: my_value
  ANOTHER_ENV_VAR: another_value
py_modules: # (Optional) A list of local directories or remote URIs that will be added to the Python path.
  - /path/to/my_module
  - s3://my_bucket/my_module
cloud: anyscale-prod # (Optional) The name of the Anyscale Cloud.
project: my-project # (Optional) The name of the Anyscale Project.
query_auth_token_enabled: True # (Optional) Whether or not to gate this service behind an auth token.
http_options:
  request_timeout_s: 60 # (Optional) Timeout for HTTP requests in seconds. Default is no timeout.
  keep_alive_timeout_s: 60 # (Optional) Timeout for HTTP keep-alive connections in seconds. Default is 5 seconds.
grpc_options:
  port: 50051 # (Optional) Port to listen on for gRPC requests. Default is 9000.
  grpc_servicer_functions: # (Optional) List of functions to expose as gRPC servicers.
    - my_module:my_function
logging_config: # (Optional) Configuration options for logging.
  encoding: JSON # JSON or TEXT.
  log_level: INFO
  enable_access_log: true
tracing_config: # (Optional) Configuration options for tracing.
  enabled: true
  exporter_import_path: my_module:custom_tracing_exporter
  sampling_ratio: 1.0
"""

    # Override the `name` field from `WorkloadConfig` so we can document it separately for jobs and services.
    name: Optional[str] = field(
        default=None,
        metadata={
            "docstring": "Unique name of the service. When running inside a workspace, this defaults to the name of the workspace, else it is required."
        },
    )

    applications: List[Dict[str, Any]] = field(
        default_factory=list,
        repr=False,
        metadata={
            "docstring": "List of Ray Serve applications to run. At least one application must be specified. For details, see the Ray Serve config file format documentation: https://docs.ray.io/en/latest/serve/production-guide/config.html"
        },
    )

    def _validate_applications(self, applications: List[Dict[str, Any]]):
        if not isinstance(applications, list):
            raise TypeError("'applications' must be a list.")
        if len(applications) == 0:
            raise ValueError("'applications' cannot be empty.")

        # Validate import paths.
        for app in applications:
            import_path = app.get("import_path", None)
            if not import_path:
                raise ValueError("Every application must specify an import path.")

            if not isinstance(import_path, str):
                raise TypeError(f"'import_path' must be a string, got: {import_path}")

            if (
                import_path.count(":") != 1
                or import_path.rfind(":") in {0, len(import_path) - 1}
                or import_path.rfind(".") in {0, len(import_path) - 1}
            ):
                raise ValueError(
                    f"'import_path' must be of the form: 'module.optional_submodule:app', but got: '{import_path}'."
                )

    query_auth_token_enabled: bool = field(
        default=True,
        repr=False,
        metadata={
            "docstring": "Whether or not queries to this service is gated behind an authentication token. If `True`, an auth token is generated the first time the service is deployed. You can find the token in the UI or by fetching the status of the service."
        },
    )

    def _validate_query_auth_token_enabled(self, query_auth_token_enabled: bool):
        if not isinstance(query_auth_token_enabled, bool):
            raise TypeError("'query_auth_token_enabled' must be a boolean.")

    http_options: Optional[Dict[str, Any]] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "HTTP options that will be passed to Ray Serve. See https://docs.ray.io/en/latest/serve/production-guide/config.html for supported options."
        },
    )

    def _validate_http_options(self, http_options: Optional[Dict[str, Any]]):
        """Validate the `http_options` field.

        This is passed through as part of the Ray Serve config, but some fields are
        disallowed (not valid when deploying Anyscale services).
        """
        if http_options is None:
            return
        elif not isinstance(http_options, dict):
            raise TypeError("'http_options' must be a dict.")

        banned_options = {"host", "port", "root_path"}
        banned_options_passed = {o for o in banned_options if o in http_options}
        if len(banned_options_passed) > 0:
            raise ValueError(
                "The following provided 'http_options' are not permitted "
                f"in Anyscale: {banned_options_passed}."
            )

    grpc_options: Optional[Dict[str, Any]] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "gRPC options that will be passed to Ray Serve. See https://docs.ray.io/en/latest/serve/production-guide/config.html for supported options."
        },
    )

    def _validate_grpc_options(self, grpc_options: Optional[Dict[str, Any]]):
        """Validate the `grpc_options` field.

        This is passed through as part of the Ray Serve config, but some fields are
        disallowed (not valid when deploying Anyscale services).
        """
        if grpc_options is None:
            return
        elif not isinstance(grpc_options, dict):
            raise TypeError("'grpc_options' must be a dict.")

        banned_options = {
            "port",
        }
        banned_options_passed = {o for o in banned_options if o in grpc_options}
        if len(banned_options_passed) > 0:
            raise ValueError(
                "The following provided 'grpc_options' are not permitted "
                f"in Anyscale: {banned_options_passed}."
            )

    logging_config: Optional[Dict[str, Any]] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Logging options that will be passed to Ray Serve. See https://docs.ray.io/en/latest/serve/production-guide/config.html for supported options."
        },
    )

    def _validate_logging_config(self, logging_config: Optional[Dict[str, Any]]):
        if logging_config is not None and not isinstance(logging_config, dict):
            raise TypeError("'logging_config' must be a dict.")

    ray_gcs_external_storage_config: Union[RayGCSExternalStorageConfig, Dict, None] = (
        field(
            default=None,
            repr=False,
            metadata={
                "docstring": "Configuration options for external storage for the Ray Global Control Store (GCS).",
                "customer_hosted_only": True,
            },
        )
    )

    def _validate_ray_gcs_external_storage_config(
        self,
        ray_gcs_external_storage_config: Union[RayGCSExternalStorageConfig, Dict, None],
    ) -> Optional[RayGCSExternalStorageConfig]:
        if ray_gcs_external_storage_config is None:
            return None

        if isinstance(ray_gcs_external_storage_config, dict):
            ray_gcs_external_storage_config = RayGCSExternalStorageConfig.from_dict(
                ray_gcs_external_storage_config
            )

        if not isinstance(ray_gcs_external_storage_config, RayGCSExternalStorageConfig):
            raise TypeError(
                "'ray_gcs_external_storage_config' must be a RayGCSExternalStorageConfig or corresponding dict."
            )

        return ray_gcs_external_storage_config

    tracing_config: Union[TracingConfig, Dict, None] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Configuration options for tracing.",
            "customer_hosted_only": True,
        },
    )

    def _validate_tracing_config(
        self, tracing_config: Union[TracingConfig, Dict, None],
    ) -> Optional[TracingConfig]:
        if tracing_config is None:
            return None

        if isinstance(tracing_config, dict):
            tracing_config = TracingConfig.from_dict(tracing_config)

        if not isinstance(tracing_config, TracingConfig):
            raise TypeError(
                "'tracing_config' must be a TracingConfig or corresponding dict."
            )

        return tracing_config


class ServiceState(ModelEnum):
    """Current state of a service."""

    STARTING = "STARTING"
    RUNNING = "RUNNING"
    # TODO(edoakes): UPDATING comes up while rolling out and rolling back.
    # This is very unexpected from a customer's point of view, we should fix it.
    UPDATING = "UPDATING"
    ROLLING_OUT = "ROLLING_OUT"
    ROLLING_BACK = "ROLLING_BACK"
    UNHEALTHY = "UNHEALTHY"
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"
    UNKNOWN = "UNKNOWN"
    SYSTEM_FAILURE = "SYSTEM_FAILURE"

    __docstrings__ = {
        STARTING: "The service is starting up.",
        RUNNING: "The service is running and healthy.",
        UPDATING: "One of the service versions or the networking stack is updating.",
        ROLLING_OUT: "The service is rolling out to a new canary version.",
        ROLLING_BACK: "The service is rolling back to the primary version.",
        UNHEALTHY: (
            "The service has become unhealthy due to an unexpected issue "
            "such as a node failure. In most cases these issues should "
            "automatically recover."
        ),
        TERMINATING: "The service is terminating.",
        TERMINATED: "The service is terminated.",
        UNKNOWN: (
            "The CLI/SDK received an unexpected state from the API server. "
            "In most cases, this means you need to update the CLI."
        ),
        SYSTEM_FAILURE: (
            "An unexpected error occurred in the Anyscale stack. "
            "Please reach out to customer support immediately."
        ),
    }


class ServiceVersionState(ModelEnum):
    """Current state of a service version."""

    STARTING = "STARTING"
    UPDATING = "UPDATING"
    RUNNING = "RUNNING"
    UNHEALTHY = "UNHEALTHY"
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"
    UNKNOWN = "UNKNOWN"
    SYSTEM_FAILURE = "SYSTEM_FAILURE"

    __docstrings__ = {
        STARTING: "The service version is starting up.",
        RUNNING: "The service version is running and healthy.",
        UPDATING: (
            "The application(s) on the service version are updating "
            "due to a config deploy or update."
        ),
        UNHEALTHY: (
            "The service version has become unhealthy due to an unexpected "
            "issue such as a node failure. In most cases these issues should "
            "automatically recover."
        ),
        TERMINATING: "The service version is terminating.",
        TERMINATED: "The service version is terminated.",
        UNKNOWN: (
            "The CLI/SDK received an unexpected state from the API server. "
            "In most cases, this means you need to update the CLI."
        ),
        SYSTEM_FAILURE: (
            "An unexpected error occurred in the Anyscale stack. "
            "Please reach out to customer support immediately."
        ),
    }


@dataclass(frozen=True)
class ServiceVersionStatus(ModelBase):
    """Current status of a single version of a running service."""

    __doc_py_example__ = """\
import anyscale
from anyscale.service.models import ServiceStatus, ServiceVersionStatus

status: ServiceStatus = anyscale.service.status("my-service")
primary_version_status: ServiceVersionStatus = status.primary_version
"""

    __doc_cli_example__ = """\
$ anyscale service status -n my-service
name: my-service
id: service2_uz6l8yhy2as5wrer3shzj6kh67
state: RUNNING
primary_version:
  id: 601bd56c4b
  state: RUNNING
  weight: 100
"""

    id: str = field(
        metadata={
            "docstring": "Unique ID of the service _version_ (*not* the service)."
        }
    )

    def _validate_id(self, id: str):  # noqa: A002
        if not isinstance(id, str):
            raise TypeError("'id' must be a string.")

    state: Union[ServiceVersionState, str] = field(
        metadata={"docstring": "Current state of the service version."}
    )

    def _validate_state(
        self, state: Union[ServiceVersionState, str]
    ) -> ServiceVersionState:
        if isinstance(state, str):
            # This will raise a ValueError if the state is unrecognized.
            state = ServiceVersionState(state)
        elif not isinstance(state, ServiceVersionState):
            raise TypeError("'state' must be a ServiceVersionState.")

        return state

    weight: int = field(
        metadata={
            "docstring": "Current canary weight of the service version. This is `100` if there is no ongoing rollout."
        }
    )

    def _validate_weight(self, weight: int):
        if not isinstance(weight, int):
            raise TypeError("'weight' must be an int.")

    config: Union[ServiceConfig, Dict] = field(
        repr=False, metadata={"docstring": "Configuration of this service version."}
    )

    def _validate_config(self, config: Union[ServiceConfig, Dict]) -> ServiceConfig:
        if isinstance(config, dict):
            config = ServiceConfig.from_dict(config)

        if not isinstance(config, ServiceConfig):
            raise TypeError("'config' must be a ServiceConfig or corresponding dict.")

        return config


@dataclass(frozen=True)
class ServiceStatus(ModelBase):
    """Current status of a service."""

    __doc_py_example__ = """\
import anyscale
from anyscale.service.models import ServiceStatus

status: ServiceStatus = anyscale.service.status("my-service")
"""

    __doc_cli_example__ = """\
$ anyscale service status -n my-service
name: my-service
id: service2_uz6l8yhy2as5wrer3shzj6kh67
state: RUNNING
primary_version:
  id: 601bd56c4b
  state: RUNNING
  weight: 100
"""

    name: str = field(metadata={"docstring": "Unique name of the service."})

    def _validate_name(self, name: str):
        if not isinstance(name, str):
            raise TypeError("'name' must be a string.")

    id: str = field(
        metadata={
            "docstring": "Unique ID of the service (generated when the service is first deployed)."
        }
    )

    def _validate_id(self, id: str):  # noqa: A002
        if not isinstance(id, str):
            raise TypeError("'id' must be a string.")

    state: Union[ServiceState, str] = field(
        metadata={"docstring": "Current state of the service."}
    )

    def _validate_state(self, state: Union[ServiceState, str]) -> ServiceState:
        if isinstance(state, str):
            # This will raise a ValueError if the state is unrecognized.
            state = ServiceState(state)
        elif not isinstance(state, ServiceState):
            raise TypeError("'state' must be a ServiceState.")

        return state

    query_url: str = field(
        repr=False, metadata={"docstring": "URL used to query the service."}
    )

    def _validate_query_url(self, query_url: str):
        if not isinstance(query_url, str):
            raise TypeError("'query_url' must be a string.")

    query_auth_token: Optional[str] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Query auth token required to query the service (if any). Must be passed in as a header like: `Authorization: Bearer <token>`."
        },
    )

    def _validate_query_auth_token(self, query_auth_token: Optional[str]):
        if query_auth_token is not None and not isinstance(query_auth_token, str):
            raise TypeError("'query_auth_token' must be a string.")

    primary_version: Union[ServiceVersionStatus, Dict, None] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Primary version of the service. During a rollout, this is the service version that was running prior to the deploy."
        },
    )

    def _validate_primary_version(
        self, primary_version: Union[ServiceVersionStatus, Dict, None]
    ) -> Optional[ServiceVersionStatus]:
        if primary_version is None:
            return None

        if isinstance(primary_version, dict):
            primary_version = ServiceVersionStatus.from_dict(primary_version)

        if not isinstance(primary_version, ServiceVersionStatus):
            raise TypeError(
                "'primary_version' must be a ServiceVersionStatus or corresponding dict."
            )

        return primary_version

    canary_version: Union[ServiceVersionStatus, Dict, None] = field(
        default=None,
        repr=False,
        metadata={
            "docstring": "Canary version of the service (the one that is being deployed). When there is no rollout ongoing, this will be `None`."
        },
    )

    def _validate_canary_version(
        self, canary_version: Union[ServiceVersionStatus, Dict, None]
    ) -> Optional[ServiceVersionStatus]:
        if canary_version is None:
            return None

        if isinstance(canary_version, dict):
            canary_version = ServiceVersionStatus.from_dict(canary_version)

        if not isinstance(canary_version, ServiceVersionStatus):
            raise TypeError(
                "'canary_version' must be a ServiceVersionStatus or corresponding dict."
            )

        return canary_version



from functools import partial
import os
import subprocess
import tempfile
from typing import cast, Optional, Tuple, Union

from anyscale._private.workload.workload_sdk import WorkloadSDK
from anyscale.client.openapi_client.models import (
    CreateExperimentalWorkspace,
    ExperimentalWorkspace,
)
from anyscale.client.openapi_client.models.session_state import SessionState
from anyscale.utils.runtime_env import parse_requirements_file
from anyscale.workspace.models import WorkspaceConfig, WorkspaceState


SSH_TEMPLATE = """
Host Head-Node
  HostName {head_node_ip}
  User ubuntu
  IdentityFile {key_path}
  StrictHostKeyChecking false
  IdentitiesOnly yes

Host {name}
  HostName 0.0.0.0
  ProxyJump Head-Node
  Port 5020
  User ray
  IdentityFile {key_path}
  StrictHostKeyChecking false
  IdentitiesOnly yes
"""

ANYSCALE_WORKSPACES_SSH_OPTIONS = [
    "-o",
    "StrictHostKeyChecking=no",
    "-o",
    "UserKnownHostsFile=/dev/null",
    "-o",
    "IdentitiesOnly=yes",
]


class PrivateWorkspaceSDK(WorkloadSDK):
    _POLLING_INTERVAL_SECONDS = 10.0
    _WAIT_TIMEOUT_SECONDS = 1800.0

    _BACKEND_SESSION_STATE_TO_WORKSPACE_STATE = {
        SessionState.STOPPED: WorkspaceState.TERMINATED,
        SessionState.TERMINATED: WorkspaceState.TERMINATED,
        SessionState.STARTINGUP: WorkspaceState.STARTING,
        SessionState.STARTUPERRORED: WorkspaceState.ERRORED,
        SessionState.RUNNING: WorkspaceState.RUNNING,
        SessionState.UPDATING: WorkspaceState.UPDATING,
        SessionState.UPDATINGERRORED: WorkspaceState.ERRORED,
        SessionState.STOPPING: WorkspaceState.TERMINATING,
        SessionState.TERMINATING: WorkspaceState.TERMINATING,
        SessionState.AWAITINGSTARTUP: WorkspaceState.STARTING,
        SessionState.AWAITINGFILEMOUNTS: WorkspaceState.STARTING,
        SessionState.TERMINATINGERRORED: WorkspaceState.ERRORED,
        SessionState.STOPPINGERRORED: WorkspaceState.ERRORED,
    }

    def _resolve_to_workspace_model(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> ExperimentalWorkspace:
        if name is None and id is None:
            raise ValueError("One of 'name' or 'id' must be provided.")

        if name is not None and id is not None:
            raise ValueError("Only one of 'name' or 'id' can be provided.")

        if id is not None and (cloud is not None or project is not None):
            raise ValueError("'cloud' and 'project' should only be used with 'name'.")

        model: Optional[ExperimentalWorkspace] = self.client.get_workspace(
            name=name, id=id, cloud=cloud, project=project
        )
        if model is None:
            if name is not None:
                raise ValueError(f"Workspace with name '{name}' was not found.")
            else:
                raise ValueError(f"Workspace with ID '{id}' was not found.")

        return model

    def create(self, config: WorkspaceConfig) -> str:
        if not config or config.name is None:
            raise ValueError("Workspace name must be configured")

        name = config.name

        compute_config_id, cloud_id = self.resolve_compute_config_and_cloud_id(
            compute_config=config.compute_config, cloud=config.cloud,
        )

        project_id = self.client.get_project_id(
            parent_cloud_id=cloud_id, name=config.project
        )

        build_id = None

        if config.containerfile is not None:
            build_id = self._image_sdk.build_image_from_containerfile(
                name=f"image-for-workspace-{name}",
                containerfile=self.get_containerfile_contents(config.containerfile),
                ray_version=config.ray_version,
            )
        elif config.image_uri is not None:
            build_id = self._image_sdk.registery_image(
                image_uri=config.image_uri,
                registry_login_secret=config.registry_login_secret,
                ray_version=config.ray_version,
            )

        dynamic_requirements = None
        if (
            config.requirements
            and self._image_sdk.enable_image_build_for_tracked_requirements
        ):
            requirements = (
                parse_requirements_file(config.requirements)
                if isinstance(config.requirements, str)
                else config.requirements
            )
            if requirements:
                build_id = self._image_sdk.build_image_from_requirements(
                    name=f"image-for-workspace-{name}",
                    base_build_id=self.client.get_default_build_id(),
                    requirements=requirements,
                )
        elif config.requirements:
            dynamic_requirements = (
                parse_requirements_file(config.requirements)
                if isinstance(config.requirements, str)
                else config.requirements
            )

        if build_id is None:
            build_id = self.client.get_default_build_id()

        workspace_id = self.client.create_workspace(
            model=CreateExperimentalWorkspace(
                name=name,
                project_id=project_id,
                compute_config_id=compute_config_id,
                cluster_environment_build_id=build_id,
                idle_timeout_minutes=config.idle_termination_minutes,
                cloud_id=cloud_id,
                skip_start=True,
            )
        )

        self._logger.info(f"Workspace created successfully id: {workspace_id}")

        if dynamic_requirements:
            self.client.update_workspace_dependencies_offline_only(
                workspace_id=workspace_id, requirements=dynamic_requirements
            )
            self._logger.info(f"Applied dynamic requirements to workspace id: {name}")
        if config.env_vars:
            self.client.update_workspace_env_vars_offline_only(
                workspace_id=workspace_id, env_vars=config.env_vars
            )
            self._logger.info(f"Applied environment variables to workspace id: {name}")

        return workspace_id

    def start(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )
        self.client.start_workspace(workspace_model.id)
        self.logger.info(f"Starting workspace '{workspace_model.name}'")
        return workspace_model.id

    def terminate(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )
        self.client.terminate_workspace(workspace_model.id)
        self.logger.info(f"Terminating workspace '{workspace_model.name}'")
        return workspace_model.id

    def status(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )
        status = self._get_workspace_status(workspace_model.id)
        return status

    def wait(  # noqa: PLR0912
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        timeout_s: float = _WAIT_TIMEOUT_SECONDS,
        state: Union[str, WorkspaceState] = WorkspaceState.RUNNING,
        interval_s: float = _POLLING_INTERVAL_SECONDS,
    ):
        if not isinstance(timeout_s, (int, float)):
            raise TypeError("timeout_s must be a float")
        if timeout_s <= 0:
            raise ValueError("timeout_s must be >= 0")

        if not isinstance(interval_s, (int, float)):
            raise TypeError("interval_s must be a float")
        if interval_s <= 0:
            raise ValueError("interval_s must be >= 0")

        if isinstance(state, str):
            try:
                state = WorkspaceState.validate(state)
            except KeyError:
                raise ValueError(f"Invalid state: {state}")

        if not isinstance(state, WorkspaceState):
            raise TypeError("'state' must be a WorkspaceState.")

        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )

        workspace_id_or_name = workspace_model.id or workspace_model.name
        curr_state = self._get_workspace_status(workspace_model.id)

        self.logger.info(
            f"Waiting for workspace '{workspace_id_or_name}' to reach target state {state}, currently in state: {curr_state}"
        )
        for _ in self.timer.poll(timeout_s=timeout_s, interval_s=interval_s):
            new_state = self._get_workspace_status(workspace_model.id)

            if new_state != curr_state:
                self.logger.info(
                    f"Workspace '{workspace_id_or_name}' transitioned from {curr_state} to {new_state}"
                )
                curr_state = new_state

            if curr_state == state:
                self.logger.info(
                    f"Workspace '{workspace_id_or_name}' reached target state, exiting"
                )
                break
        else:
            raise TimeoutError(
                f"Workspace '{workspace_id_or_name}' did not reach target state {state} within {timeout_s}s. Last seen state: {curr_state}."
            )

    def generate_ssh_config_file(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        ssh_config_path: Optional[str] = None,
    ) -> Tuple[str, str]:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )

        assert (
            self.status(id=workspace_model.id) == WorkspaceState.RUNNING
        ), "Workspace must be running to generate SSH config file"

        head_node_ip = self.client.get_cluster_head_node_ip(workspace_model.cluster_id)
        ssh_key = self.client.get_cluster_ssh_key(workspace_model.cluster_id)

        def _store_ssh_key(name: str, ssh_key: str, key_dir: str) -> str:
            key_path = os.path.join(key_dir, f"{name}.pem")
            os.makedirs(os.path.dirname(key_path), exist_ok=True)

            with open(key_path, "w", opener=partial(os.open, mode=0o600)) as f:
                f.write(ssh_key)

            return key_path

        if ssh_config_path is None:
            ssh_config_path = tempfile.mkdtemp()

        key_path = _store_ssh_key(
            ssh_key.key_name, ssh_key.private_key, ssh_config_path
        )

        ssh_config = SSH_TEMPLATE.format(
            head_node_ip=head_node_ip, key_path=key_path, name=workspace_model.name
        )

        config_file_name = os.path.join(ssh_config_path, "config")
        with open(config_file_name, "w") as f:
            f.write(ssh_config)

        return workspace_model.name, config_file_name

    def run_command(
        self,
        command: str,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )
        host_name, config_file = self.generate_ssh_config_file(id=workspace_model.id)
        return subprocess.run(
            ["ssh"]
            + ANYSCALE_WORKSPACES_SSH_OPTIONS
            + ["-F", config_file, host_name, command],
            **kwargs,
        )

    def get_default_dir_name(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        workspace_model = self._resolve_to_workspace_model(
            id=id, name=name, cloud=cloud, project=project
        )
        return self.client.get_workspace_default_dir_name(workspace_model.id)

    def _get_workspace_status(self, workspace_id: str) -> WorkspaceState:
        workspace_session_state = self.client.get_workspace_status(workspace_id)
        return cast(
            WorkspaceState,
            self._BACKEND_SESSION_STATE_TO_WORKSPACE_STATE.get(
                workspace_session_state, WorkspaceState.UNKNOWN
            ),
        )



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class ProductionJobConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'entrypoint': 'str',
        'ray_serve_config': 'object',
        'runtime_env': 'RayRuntimeEnvConfig',
        'build_id': 'str',
        'compute_config_id': 'str',
        'compute_config': 'CreateClusterComputeConfig',
        'max_retries': 'int',
        'runtime_env_config': 'RayRuntimeEnvConfig'
    }

    attribute_map = {
        'entrypoint': 'entrypoint',
        'ray_serve_config': 'ray_serve_config',
        'runtime_env': 'runtime_env',
        'build_id': 'build_id',
        'compute_config_id': 'compute_config_id',
        'compute_config': 'compute_config',
        'max_retries': 'max_retries',
        'runtime_env_config': 'runtime_env_config'
    }

    def __init__(self, entrypoint='', ray_serve_config=None, runtime_env=None, build_id=None, compute_config_id=None, compute_config=None, max_retries=5, runtime_env_config=None, local_vars_configuration=None):  # noqa: E501
        """ProductionJobConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._entrypoint = None
        self._ray_serve_config = None
        self._runtime_env = None
        self._build_id = None
        self._compute_config_id = None
        self._compute_config = None
        self._max_retries = None
        self._runtime_env_config = None
        self.discriminator = None

        if entrypoint is not None:
            self.entrypoint = entrypoint
        if ray_serve_config is not None:
            self.ray_serve_config = ray_serve_config
        if runtime_env is not None:
            self.runtime_env = runtime_env
        self.build_id = build_id
        self.compute_config_id = compute_config_id
        if compute_config is not None:
            self.compute_config = compute_config
        if max_retries is not None:
            self.max_retries = max_retries
        if runtime_env_config is not None:
            self.runtime_env_config = runtime_env_config

    @property
    def entrypoint(self):
        """Gets the entrypoint of this ProductionJobConfig.  # noqa: E501

        A script that will be run to start your job.This command will be run in the root directory of the specified runtime env. Eg. 'python script.py'  # noqa: E501

        :return: The entrypoint of this ProductionJobConfig.  # noqa: E501
        :rtype: str
        """
        return self._entrypoint

    @entrypoint.setter
    def entrypoint(self, entrypoint):
        """Sets the entrypoint of this ProductionJobConfig.

        A script that will be run to start your job.This command will be run in the root directory of the specified runtime env. Eg. 'python script.py'  # noqa: E501

        :param entrypoint: The entrypoint of this ProductionJobConfig.  # noqa: E501
        :type: str
        """

        self._entrypoint = entrypoint

    @property
    def ray_serve_config(self):
        """Gets the ray_serve_config of this ProductionJobConfig.  # noqa: E501

        The Ray Serve config to use for this Production service. This config defines your Ray Serve application, and will be passed directly to Ray Serve. You can learn more about Ray Serve config files here: https://docs.ray.io/en/latest/serve/production-guide/config.html  # noqa: E501

        :return: The ray_serve_config of this ProductionJobConfig.  # noqa: E501
        :rtype: object
        """
        return self._ray_serve_config

    @ray_serve_config.setter
    def ray_serve_config(self, ray_serve_config):
        """Sets the ray_serve_config of this ProductionJobConfig.

        The Ray Serve config to use for this Production service. This config defines your Ray Serve application, and will be passed directly to Ray Serve. You can learn more about Ray Serve config files here: https://docs.ray.io/en/latest/serve/production-guide/config.html  # noqa: E501

        :param ray_serve_config: The ray_serve_config of this ProductionJobConfig.  # noqa: E501
        :type: object
        """

        self._ray_serve_config = ray_serve_config

    @property
    def runtime_env(self):
        """Gets the runtime_env of this ProductionJobConfig.  # noqa: E501

        A ray runtime env json. Your entrypoint will be run in the environment specified by this runtime env.  # noqa: E501

        :return: The runtime_env of this ProductionJobConfig.  # noqa: E501
        :rtype: RayRuntimeEnvConfig
        """
        return self._runtime_env

    @runtime_env.setter
    def runtime_env(self, runtime_env):
        """Sets the runtime_env of this ProductionJobConfig.

        A ray runtime env json. Your entrypoint will be run in the environment specified by this runtime env.  # noqa: E501

        :param runtime_env: The runtime_env of this ProductionJobConfig.  # noqa: E501
        :type: RayRuntimeEnvConfig
        """

        self._runtime_env = runtime_env

    @property
    def build_id(self):
        """Gets the build_id of this ProductionJobConfig.  # noqa: E501

        The id of the cluster env build. This id will determine the docker image your job is run on.  # noqa: E501

        :return: The build_id of this ProductionJobConfig.  # noqa: E501
        :rtype: str
        """
        return self._build_id

    @build_id.setter
    def build_id(self, build_id):
        """Sets the build_id of this ProductionJobConfig.

        The id of the cluster env build. This id will determine the docker image your job is run on.  # noqa: E501

        :param build_id: The build_id of this ProductionJobConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and build_id is None:  # noqa: E501
            raise ValueError("Invalid value for `build_id`, must not be `None`")  # noqa: E501

        self._build_id = build_id

    @property
    def compute_config_id(self):
        """Gets the compute_config_id of this ProductionJobConfig.  # noqa: E501

        The id of the compute configuration that you want to use. This id will specify the resources required for your job  # noqa: E501

        :return: The compute_config_id of this ProductionJobConfig.  # noqa: E501
        :rtype: str
        """
        return self._compute_config_id

    @compute_config_id.setter
    def compute_config_id(self, compute_config_id):
        """Sets the compute_config_id of this ProductionJobConfig.

        The id of the compute configuration that you want to use. This id will specify the resources required for your job  # noqa: E501

        :param compute_config_id: The compute_config_id of this ProductionJobConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and compute_config_id is None:  # noqa: E501
            raise ValueError("Invalid value for `compute_config_id`, must not be `None`")  # noqa: E501

        self._compute_config_id = compute_config_id

    @property
    def compute_config(self):
        """Gets the compute_config of this ProductionJobConfig.  # noqa: E501

        One-off compute that the cluster will use.  # noqa: E501

        :return: The compute_config of this ProductionJobConfig.  # noqa: E501
        :rtype: CreateClusterComputeConfig
        """
        return self._compute_config

    @compute_config.setter
    def compute_config(self, compute_config):
        """Sets the compute_config of this ProductionJobConfig.

        One-off compute that the cluster will use.  # noqa: E501

        :param compute_config: The compute_config of this ProductionJobConfig.  # noqa: E501
        :type: CreateClusterComputeConfig
        """

        self._compute_config = compute_config

    @property
    def max_retries(self):
        """Gets the max_retries of this ProductionJobConfig.  # noqa: E501

        The number of retries this job will attempt on failure. Set to None to set infinite retries  # noqa: E501

        :return: The max_retries of this ProductionJobConfig.  # noqa: E501
        :rtype: int
        """
        return self._max_retries

    @max_retries.setter
    def max_retries(self, max_retries):
        """Sets the max_retries of this ProductionJobConfig.

        The number of retries this job will attempt on failure. Set to None to set infinite retries  # noqa: E501

        :param max_retries: The max_retries of this ProductionJobConfig.  # noqa: E501
        :type: int
        """

        self._max_retries = max_retries

    @property
    def runtime_env_config(self):
        """Gets the runtime_env_config of this ProductionJobConfig.  # noqa: E501

        DEPRECATED: Use runtime_env  # noqa: E501

        :return: The runtime_env_config of this ProductionJobConfig.  # noqa: E501
        :rtype: RayRuntimeEnvConfig
        """
        return self._runtime_env_config

    @runtime_env_config.setter
    def runtime_env_config(self, runtime_env_config):
        """Sets the runtime_env_config of this ProductionJobConfig.

        DEPRECATED: Use runtime_env  # noqa: E501

        :param runtime_env_config: The runtime_env_config of this ProductionJobConfig.  # noqa: E501
        :type: RayRuntimeEnvConfig
        """

        self._runtime_env_config = runtime_env_config

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ProductionJobConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ProductionJobConfig):
            return True

        return self.to_dict() != other.to_dict()



from typing import Any, Optional

from anyscale._private.anyscale_client import AnyscaleClientInterface
from anyscale._private.sdk import sdk_docs
from anyscale._private.sdk.base_sdk import BaseSDK, Timer
from anyscale.cli_logger import BlockLogger
from anyscale.llm.models._private.models_sdk import PrivateLLMModelsSDK
from anyscale.llm.models.commands import (
    _RETRIEVE_ARG_DOCSTRINGS,
    _RETRIEVE_EXAMPLE,
)


class LLMModelsSDK(BaseSDK):
    def __init__(
        self,
        *,
        logger: Optional[BlockLogger] = None,
        client: Optional[AnyscaleClientInterface] = None,
        timer: Optional[Timer] = None,
    ):
        self._private_sdk = PrivateLLMModelsSDK(
            logger=logger, client=client, timer=timer
        )

    @sdk_docs(
        doc_py_example=_RETRIEVE_EXAMPLE, arg_docstrings=_RETRIEVE_ARG_DOCSTRINGS,
    )
    def retrieve(self, model_id: str) -> Any:
        """Retrieves model card for a finetuned model."""
        return self._private_sdk.retrieve(model_id)



from typing import Optional, Tuple, Union

from anyscale._private.anyscale_client import AnyscaleClientInterface
from anyscale._private.sdk import sdk_docs
from anyscale._private.sdk.base_sdk import Timer
from anyscale.cli_logger import BlockLogger
from anyscale.workspace._private.workspace_sdk import PrivateWorkspaceSDK
from anyscale.workspace.commands import (
    _CREATE_ARG_DOCSTRINGS,
    _CREATE_EXAMPLE,
    _GENERATE_SSH_CONFIG_FILE_ARG_DOCSTRINGS,
    _GENERATE_SSH_CONFIG_FILE_EXAMPLE,
    _RUN_COMMAND_ARG_DOCSTRINGS,
    _RUN_COMMAND_EXAMPLE,
    _START_ARG_DOCSTRINGS,
    _START_EXAMPLE,
    _STATUS_ARG_DOCSTRINGS,
    _STATUS_EXAMPLE,
    _TERMINATE_ARG_DOCSTRINGS,
    _TERMINATE_EXAMPLE,
    _WAIT_ARG_DOCSTRINGS,
    _WAIT_EXAMPLE,
    create,
    generate_ssh_config_file,
    run_command,
    start,
    status,
    terminate,
    wait,
)
from anyscale.workspace.models import WorkspaceConfig, WorkspaceState


class WorkspaceSDK:
    def __init__(
        self,
        *,
        client: Optional[AnyscaleClientInterface] = None,
        logger: Optional[BlockLogger] = None,
        timer: Optional[Timer] = None,
    ):
        self._private_sdk = PrivateWorkspaceSDK(
            client=client, logger=logger, timer=timer
        )

    @sdk_docs(
        doc_py_example=_CREATE_EXAMPLE, arg_docstrings=_CREATE_ARG_DOCSTRINGS,
    )
    def create(self, config: WorkspaceConfig) -> str:  # noqa: F811
        """Create a workspace.

        Returns the id of the workspace.
        """
        return self._private_sdk.create(config=config)

    @sdk_docs(
        doc_py_example=_START_EXAMPLE, arg_docstrings=_START_ARG_DOCSTRINGS,
    )
    def start(  # noqa: F811
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        """Start a workspace.

        Returns the id of the started workspace.
        """
        return self._private_sdk.start(name=name, id=id, cloud=cloud, project=project)

    @sdk_docs(
        doc_py_example=_TERMINATE_EXAMPLE, arg_docstrings=_TERMINATE_ARG_DOCSTRINGS,
    )
    def terminate(  # noqa: F811
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        """Terminate a workspace.

        Returns the id of the terminated workspace.
        """
        return self._private_sdk.terminate(
            name=name, id=id, cloud=cloud, project=project
        )

    @sdk_docs(
        doc_py_example=_STATUS_EXAMPLE, arg_docstrings=_STATUS_ARG_DOCSTRINGS,
    )
    def status(  # noqa: F811
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        """Get the status of a workspace.

        Returns the status of the workspace.
        """
        return self._private_sdk.status(name=name, id=id, cloud=cloud, project=project)

    @sdk_docs(
        doc_py_example=_WAIT_EXAMPLE, arg_docstrings=_WAIT_ARG_DOCSTRINGS,
    )
    def wait(  # noqa: F811
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        timeout_s: float = 1800,
        state: Union[str, WorkspaceState] = WorkspaceState.RUNNING,
    ) -> str:
        """Wait for a workspace to reach a terminal state.

        Returns the id of the workspace.
        """
        return self._private_sdk.wait(
            name=name,
            id=id,
            cloud=cloud,
            project=project,
            timeout_s=timeout_s,
            state=state,
        )

    @sdk_docs(
        doc_py_example=_GENERATE_SSH_CONFIG_FILE_EXAMPLE,
        arg_docstrings=_GENERATE_SSH_CONFIG_FILE_ARG_DOCSTRINGS,
    )
    def generate_ssh_config_file(  # noqa: F811
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        ssh_config_path: Optional[str] = None,
    ) -> Tuple[str, str]:
        """Generate an SSH config file for a workspace.

        Returns a tuple of host name and config file path.
        """
        return self._private_sdk.generate_ssh_config_file(
            name=name,
            id=id,
            cloud=cloud,
            project=project,
            ssh_config_path=ssh_config_path,
        )

    @sdk_docs(
        doc_py_example=_RUN_COMMAND_EXAMPLE, arg_docstrings=_RUN_COMMAND_ARG_DOCSTRINGS,
    )
    def run_command(  # noqa: F811
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        command: str,
        **kwargs,
    ):
        """Run a command on a workspace.

        Returns the output of the command.
        """
        return self._private_sdk.run_command(
            name=name, id=id, cloud=cloud, project=project, command=command, **kwargs
        )



from contextlib import redirect_stderr, redirect_stdout
import io
import os
from unittest.mock import Mock, patch

from botocore.exceptions import ClientError
import click
from google.api_core.exceptions import GoogleAPICallError
from googleapiclient.errors import HttpError
import pytest

from anyscale.authenticate import get_auth_api_client
from anyscale.cli_logger import BlockLogger, CloudSetupLogger
from anyscale.client.openapi_client.models import CloudAnalyticsEventCloudResource
from anyscale.client.openapi_client.rest import ApiException as ApiExceptionInternal
from anyscale.sdk.anyscale_client.rest import ApiException as ApiExceptionExternal


def test_warning_to_stderr():
    # Test warnings are printed to stderr
    log = BlockLogger()
    f = io.StringIO()
    warning_message = "test_warning"
    with redirect_stderr(f):
        log.warning(warning_message)

    assert warning_message in f.getvalue()


def test_block():
    log = BlockLogger()

    # Test opening block, writing info, closing block
    log.open_block(block_label="block1")
    log.info("msg1", block_label="block1")
    log.close_block(block_label="block1")

    # Test error raised when opening two blocks without closing first
    log.open_block(block_label="block1")
    with pytest.raises(Exception):  # noqa: PT011
        log.open_block(block_label="block2")

    # Test error raised when closing block that wasn't opened
    with pytest.raises(Exception):  # noqa: PT011
        log.close_block(block_label="block2")

    # Test error raised when writing to block that wasn't opened
    with pytest.raises(Exception):  # noqa: PT011
        log.info("msg2", block_label="block2")


def test_debug_env_var():
    log = BlockLogger()
    f = io.StringIO()
    debug_message = "test_debug"

    # Test debug message is not printed to stdout if ANYSCALE_DEBUG != 1
    os.environ.pop("ANYSCALE_DEBUG", None)
    with redirect_stdout(f):
        log.debug(debug_message)
    assert f.getvalue() == ""

    # Test debug message is printed to stdout if ANYSCALE_DEBUG == 1
    os.environ["ANYSCALE_DEBUG"] = "1"
    with redirect_stdout(f):
        log.debug(debug_message)
    assert debug_message in f.getvalue()


def test_format_api_exception_internal():
    # Tests that API exceptions are correctly formatted for the internal API
    BlockLogger()
    with patch.multiple(
        "anyscale.authenticate.AuthenticationBlock",
        _validate_api_client_auth=Mock(),
        _validate_credentials_format=Mock(),
    ):
        mock_api_client = get_auth_api_client(
            cli_token="fake_cli_token", host="fake_host"
        ).api_client

    # Test non ApiExceptions are not caught by log.format_api_exception
    with patch.multiple(
        "anyscale.api.openapi_client.ApiClient",
        call_api=Mock(side_effect=ZeroDivisionError()),
    ), pytest.raises(ZeroDivisionError):
        mock_api_client.get_project_api_v2_projects_project_id_get("bad_project_id")

    e = ApiExceptionInternal()
    e.headers = Mock(_container={})
    with patch.multiple(
        "anyscale.api.openapi_client.ApiClient", call_api=Mock(side_effect=e),
    ):
        # Test original ApiException is raised if ANYSCALE_DEBUG == 1
        os.environ["ANYSCALE_DEBUG"] = "1"
        with pytest.raises(ApiExceptionInternal):
            mock_api_client.get_project_api_v2_projects_project_id_get("bad_project_id")

        # Test formatted ClickException is raised if ANYSCALE_DEBUG != 1
        os.environ.pop("ANYSCALE_DEBUG", None)
        with pytest.raises(click.ClickException):
            mock_api_client.get_project_api_v2_projects_project_id_get("bad_project_id")


def test_format_api_exception_external():
    # Tests that API exceptions are correctly formatted for the external API
    BlockLogger()
    with patch.multiple(
        "anyscale.authenticate.AuthenticationBlock",
        _validate_api_client_auth=Mock(),
        _validate_credentials_format=Mock(),
    ):
        mock_api_client = get_auth_api_client(
            cli_token="fake_cli_token", host="fake_host"
        ).anyscale_api_client

    # Test non ApiExceptions are not caught by log.format_api_exception
    with patch.multiple(
        "anyscale.sdk.anyscale_client.ApiClient",
        call_api=Mock(side_effect=ZeroDivisionError()),
    ), pytest.raises(ZeroDivisionError):
        mock_api_client.list_projects()

    e = ApiExceptionExternal()
    e.headers = Mock(_container={})
    with patch.multiple(
        "anyscale.sdk.anyscale_client.ApiClient", call_api=Mock(side_effect=e),
    ):
        # Test original ApiException is raised if ANYSCALE_DEBUG == 1
        os.environ["ANYSCALE_DEBUG"] = "1"
        with pytest.raises(ApiExceptionExternal):
            mock_api_client.list_projects()

        # Test formatted ClickException is raised if ANYSCALE_DEBUG != 1
        os.environ.pop("ANYSCALE_DEBUG", None)
        with pytest.raises(click.ClickException):
            mock_api_client.list_projects()


@pytest.mark.parametrize(
    "exception_class", [Exception, GoogleAPICallError, HttpError, ClientError,]
)
@pytest.mark.parametrize("code_exists", [True, False])
@pytest.mark.parametrize("reason_exists", [True, False])
@pytest.mark.parametrize("valid_error_details", [True, False])
def test_log_resource_exception(
    exception_class, code_exists, reason_exists, valid_error_details: bool
):
    logger = CloudSetupLogger()
    mock_log_error = Mock()
    mock_cloud_resource = Mock()
    mock_exception = Mock(spec=exception_class)
    mock_error_reason = "mock_error_reason" if reason_exists else None
    mock_error_code = 404 if code_exists else None
    if exception_class == ClientError:
        mock_exception.response = {
            "Error": {"Code": mock_error_reason},
            "ResponseMetadata": {"HTTPStatusCode": mock_error_code},
        }
    elif exception_class == GoogleAPICallError:
        mock_exception.code = mock_error_code
        mock_exception.reason = mock_error_reason
    elif exception_class == HttpError:
        mock_exception.status_code = mock_error_code
        mock_exception.error_details = (
            [{"reason": mock_error_reason}] if valid_error_details else "wrong_format"
        )
    with patch.multiple(
        "anyscale.cli_logger.CloudSetupLogger", log_resource_error=mock_log_error,
    ):
        if exception_class in (GoogleAPICallError, HttpError):
            logger.log_resource_exception(
                CloudAnalyticsEventCloudResource.GCP_PROJECT, mock_exception
            )
            if exception_class == GoogleAPICallError:
                mock_log_error.assert_called_once_with(
                    CloudAnalyticsEventCloudResource.GCP_PROJECT,
                    mock_error_reason,
                    mock_error_code,
                    "GoogleAPICallError",
                )
            else:
                expected_error_reason = (
                    mock_error_reason if valid_error_details else None
                )
                mock_log_error.assert_called_once_with(
                    CloudAnalyticsEventCloudResource.GCP_PROJECT,
                    expected_error_reason,
                    mock_error_code,
                    "HttpError",
                )
        else:
            logger.log_resource_exception(mock_cloud_resource, mock_exception)
            if exception_class == ClientError:
                mock_log_error.assert_called_once_with(
                    mock_cloud_resource,
                    mock_error_reason,
                    mock_error_code,
                    "ClientError",
                )
            else:
                mock_log_error.assert_called_once_with(
                    mock_cloud_resource,
                    None,
                    None,
                    f"UnknownExceptionType_{Exception.__name__}",
                )


@pytest.mark.parametrize("error_reason", ["not_found", None])
@pytest.mark.parametrize("status_code", [404, None])
@pytest.mark.parametrize("unhandled_exception", ["MockException", None])
def test_log_resource_error(error_reason, status_code, unhandled_exception):
    logger = CloudSetupLogger()
    mock_cloud_resource = Mock()
    logger.log_resource_error(
        mock_cloud_resource, error_reason, status_code, unhandled_exception
    )
    assert len(logger.cloud_resource_errors) == 1
    cloud_resource_error = logger.cloud_resource_errors[0]
    assert cloud_resource_error.cloud_resource == mock_cloud_resource
    expected_error_code = f"{error_reason if error_reason else 'unknown'}"
    if status_code:
        expected_error_code = f"{expected_error_code},{status_code}"
    if unhandled_exception:
        expected_error_code = f"{expected_error_code},{unhandled_exception}"
    assert cloud_resource_error.error_code == expected_error_code

    # Log duplicated error
    logger.log_resource_error(
        mock_cloud_resource, error_reason, status_code, unhandled_exception
    )
    assert len(logger.cloud_resource_errors) == 1


def test_get_cloud_provider_errors():
    logger = CloudSetupLogger()
    mock_cloud_resource = Mock()
    logger.log_resource_error(mock_cloud_resource, "not_found", 404)
    errors = logger.get_cloud_provider_errors()
    assert len(errors) == 1
    assert errors[0].cloud_resource == mock_cloud_resource
    assert errors[0].error_code == "not_found,404"


def test_clear_cloud_provider_errors():
    logger = CloudSetupLogger()
    mock_cloud_resource = Mock()
    logger.log_resource_error(mock_cloud_resource, "not_found", 404)
    logger.clear_cloud_provider_errors()
    assert len(logger.cloud_resource_errors) == 0



from anyscale._private.sdk import sdk_command
from anyscale.client.openapi_client import FineTunedModel
from anyscale.llm.models._private.models_sdk import PrivateLLMModelsSDK


_LLM_MODELS_SDK_SINGLETON_KEY = "llm_models_sdk"


_RETRIEVE_EXAMPLE = """
import anyscale

anyscale.llm.models.retrieve(model_id="my-model-id")
"""

_RETRIEVE_ARG_DOCSTRINGS = {
    "model_id": " ID of the finetuned model for which the model card is being retrieved."
}


@sdk_command(
    _LLM_MODELS_SDK_SINGLETON_KEY,
    PrivateLLMModelsSDK,
    doc_py_example=_RETRIEVE_EXAMPLE,
    arg_docstrings=_RETRIEVE_ARG_DOCSTRINGS,
)
def retrieve(model_id: str, _sdk: PrivateLLMModelsSDK,) -> FineTunedModel:
    """Retrieves model card for a finetuned model."""
    return _sdk.retrieve(model_id)



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class UXInstance(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cpu_count': 'int',
        'memory_in_gib': 'float',
        'gpu_type': 'str',
        'gpu_count': 'int',
        'cloud_type': 'CloudType',
        'support_type': 'str',
        'name': 'str',
        'family': 'str',
        'group_key': 'str',
        'anyscale_credits_per_hour': 'float',
        'available_in_anyscale_hosted': 'bool'
    }

    attribute_map = {
        'cpu_count': 'cpu_count',
        'memory_in_gib': 'memory_in_gib',
        'gpu_type': 'gpu_type',
        'gpu_count': 'gpu_count',
        'cloud_type': 'cloud_type',
        'support_type': 'support_type',
        'name': 'name',
        'family': 'family',
        'group_key': 'group_key',
        'anyscale_credits_per_hour': 'anyscale_credits_per_hour',
        'available_in_anyscale_hosted': 'available_in_anyscale_hosted'
    }

    def __init__(self, cpu_count=None, memory_in_gib=None, gpu_type=None, gpu_count=None, cloud_type=None, support_type=None, name=None, family=None, group_key=None, anyscale_credits_per_hour=None, available_in_anyscale_hosted=None, local_vars_configuration=None):  # noqa: E501
        """UXInstance - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._cpu_count = None
        self._memory_in_gib = None
        self._gpu_type = None
        self._gpu_count = None
        self._cloud_type = None
        self._support_type = None
        self._name = None
        self._family = None
        self._group_key = None
        self._anyscale_credits_per_hour = None
        self._available_in_anyscale_hosted = None
        self.discriminator = None

        self.cpu_count = cpu_count
        self.memory_in_gib = memory_in_gib
        if gpu_type is not None:
            self.gpu_type = gpu_type
        if gpu_count is not None:
            self.gpu_count = gpu_count
        self.cloud_type = cloud_type
        self.support_type = support_type
        self.name = name
        self.family = family
        self.group_key = group_key
        self.anyscale_credits_per_hour = anyscale_credits_per_hour
        self.available_in_anyscale_hosted = available_in_anyscale_hosted

    @property
    def cpu_count(self):
        """Gets the cpu_count of this UXInstance.  # noqa: E501

        The number of CPUs  # noqa: E501

        :return: The cpu_count of this UXInstance.  # noqa: E501
        :rtype: int
        """
        return self._cpu_count

    @cpu_count.setter
    def cpu_count(self, cpu_count):
        """Sets the cpu_count of this UXInstance.

        The number of CPUs  # noqa: E501

        :param cpu_count: The cpu_count of this UXInstance.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and cpu_count is None:  # noqa: E501
            raise ValueError("Invalid value for `cpu_count`, must not be `None`")  # noqa: E501

        self._cpu_count = cpu_count

    @property
    def memory_in_gib(self):
        """Gets the memory_in_gib of this UXInstance.  # noqa: E501

        The amount of memory in GiB  # noqa: E501

        :return: The memory_in_gib of this UXInstance.  # noqa: E501
        :rtype: float
        """
        return self._memory_in_gib

    @memory_in_gib.setter
    def memory_in_gib(self, memory_in_gib):
        """Sets the memory_in_gib of this UXInstance.

        The amount of memory in GiB  # noqa: E501

        :param memory_in_gib: The memory_in_gib of this UXInstance.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and memory_in_gib is None:  # noqa: E501
            raise ValueError("Invalid value for `memory_in_gib`, must not be `None`")  # noqa: E501

        self._memory_in_gib = memory_in_gib

    @property
    def gpu_type(self):
        """Gets the gpu_type of this UXInstance.  # noqa: E501

        The type of GPU  # noqa: E501

        :return: The gpu_type of this UXInstance.  # noqa: E501
        :rtype: str
        """
        return self._gpu_type

    @gpu_type.setter
    def gpu_type(self, gpu_type):
        """Sets the gpu_type of this UXInstance.

        The type of GPU  # noqa: E501

        :param gpu_type: The gpu_type of this UXInstance.  # noqa: E501
        :type: str
        """

        self._gpu_type = gpu_type

    @property
    def gpu_count(self):
        """Gets the gpu_count of this UXInstance.  # noqa: E501

        The number of GPUs  # noqa: E501

        :return: The gpu_count of this UXInstance.  # noqa: E501
        :rtype: int
        """
        return self._gpu_count

    @gpu_count.setter
    def gpu_count(self, gpu_count):
        """Sets the gpu_count of this UXInstance.

        The number of GPUs  # noqa: E501

        :param gpu_count: The gpu_count of this UXInstance.  # noqa: E501
        :type: int
        """

        self._gpu_count = gpu_count

    @property
    def cloud_type(self):
        """Gets the cloud_type of this UXInstance.  # noqa: E501

        The cloud provider  # noqa: E501

        :return: The cloud_type of this UXInstance.  # noqa: E501
        :rtype: CloudType
        """
        return self._cloud_type

    @cloud_type.setter
    def cloud_type(self, cloud_type):
        """Sets the cloud_type of this UXInstance.

        The cloud provider  # noqa: E501

        :param cloud_type: The cloud_type of this UXInstance.  # noqa: E501
        :type: CloudType
        """
        if self.local_vars_configuration.client_side_validation and cloud_type is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_type`, must not be `None`")  # noqa: E501

        self._cloud_type = cloud_type

    @property
    def support_type(self):
        """Gets the support_type of this UXInstance.  # noqa: E501

        The instance should be shown in the UI and through API (SDK/CLI).  # noqa: E501

        :return: The support_type of this UXInstance.  # noqa: E501
        :rtype: str
        """
        return self._support_type

    @support_type.setter
    def support_type(self, support_type):
        """Sets the support_type of this UXInstance.

        The instance should be shown in the UI and through API (SDK/CLI).  # noqa: E501

        :param support_type: The support_type of this UXInstance.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and support_type is None:  # noqa: E501
            raise ValueError("Invalid value for `support_type`, must not be `None`")  # noqa: E501

        self._support_type = support_type

    @property
    def name(self):
        """Gets the name of this UXInstance.  # noqa: E501

        Name of this instance type.  # noqa: E501

        :return: The name of this UXInstance.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this UXInstance.

        Name of this instance type.  # noqa: E501

        :param name: The name of this UXInstance.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def family(self):
        """Gets the family of this UXInstance.  # noqa: E501

        Family of this instance type.  # noqa: E501

        :return: The family of this UXInstance.  # noqa: E501
        :rtype: str
        """
        return self._family

    @family.setter
    def family(self, family):
        """Sets the family of this UXInstance.

        Family of this instance type.  # noqa: E501

        :param family: The family of this UXInstance.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and family is None:  # noqa: E501
            raise ValueError("Invalid value for `family`, must not be `None`")  # noqa: E501

        self._family = family

    @property
    def group_key(self):
        """Gets the group_key of this UXInstance.  # noqa: E501

        The groupKey indicates which instance types are similar and should be grouped together  # noqa: E501

        :return: The group_key of this UXInstance.  # noqa: E501
        :rtype: str
        """
        return self._group_key

    @group_key.setter
    def group_key(self, group_key):
        """Sets the group_key of this UXInstance.

        The groupKey indicates which instance types are similar and should be grouped together  # noqa: E501

        :param group_key: The group_key of this UXInstance.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and group_key is None:  # noqa: E501
            raise ValueError("Invalid value for `group_key`, must not be `None`")  # noqa: E501

        self._group_key = group_key

    @property
    def anyscale_credits_per_hour(self):
        """Gets the anyscale_credits_per_hour of this UXInstance.  # noqa: E501

        Anyscale credits per hour for the instance.  # noqa: E501

        :return: The anyscale_credits_per_hour of this UXInstance.  # noqa: E501
        :rtype: float
        """
        return self._anyscale_credits_per_hour

    @anyscale_credits_per_hour.setter
    def anyscale_credits_per_hour(self, anyscale_credits_per_hour):
        """Sets the anyscale_credits_per_hour of this UXInstance.

        Anyscale credits per hour for the instance.  # noqa: E501

        :param anyscale_credits_per_hour: The anyscale_credits_per_hour of this UXInstance.  # noqa: E501
        :type: float
        """
        if self.local_vars_configuration.client_side_validation and anyscale_credits_per_hour is None:  # noqa: E501
            raise ValueError("Invalid value for `anyscale_credits_per_hour`, must not be `None`")  # noqa: E501

        self._anyscale_credits_per_hour = anyscale_credits_per_hour

    @property
    def available_in_anyscale_hosted(self):
        """Gets the available_in_anyscale_hosted of this UXInstance.  # noqa: E501

        If the instance type is available for anyscale hosted.  # noqa: E501

        :return: The available_in_anyscale_hosted of this UXInstance.  # noqa: E501
        :rtype: bool
        """
        return self._available_in_anyscale_hosted

    @available_in_anyscale_hosted.setter
    def available_in_anyscale_hosted(self, available_in_anyscale_hosted):
        """Sets the available_in_anyscale_hosted of this UXInstance.

        If the instance type is available for anyscale hosted.  # noqa: E501

        :param available_in_anyscale_hosted: The available_in_anyscale_hosted of this UXInstance.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and available_in_anyscale_hosted is None:  # noqa: E501
            raise ValueError("Invalid value for `available_in_anyscale_hosted`, must not be `None`")  # noqa: E501

        self._available_in_anyscale_hosted = available_in_anyscale_hosted

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UXInstance):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, UXInstance):
            return True

        return self.to_dict() != other.to_dict()
